/*
 * File:	clips2c.cc
 * Purpose:	Clips subset reading and manipulation
 * Author:	Julian Smart
 * Created:	1993
 * Updated:	
 * Copyright:	(c) 1993, AIAI, University of Edinburgh
 */

static const char sccsid[] = "%W% %G%";

// #include <wx.h>
#include <stdarg.h>
#include <ctype.h>
#include <iostream.h>

#include "expr.h"
#include "wx_hash.h"
#include "clips2c.h"

extern "C" void add_expr(char *);
extern "C" void LexFromFile(FILE *fd);
extern "C" void LexFromString(char *buf);

extern long NewId(void);
extern void Unix2DosFilename(char *);

void ClipsNameToC(char *name);

ClipsDatabase *hyClipsDatabase = NULL;
ClipsErrorHandler CurrentClipsErrorHandler;

NameSpace TheNameSpace;
FunctionTable TheFunctionTable;

wxList ConstructList;

static int generationMode = MODE_PLAIN_CLIPS;

static wxNode *PositionToInsert = NULL;

void SetGenerationMode(int mode)
{
  generationMode = mode;
}

int GetGenerationMode(void)
{
  return generationMode;
}

// Clips 'database' (list of expressions)
ClipsDatabase::ClipsDatabase(ClipsErrorHandler handler)
{
  position = NULL;
  hash_table = NULL;
  attribute_to_hash = NULL;
  CurrentClipsErrorHandler = handler;
  noErrors = 0;
}

ClipsDatabase::ClipsDatabase(ClipsType type, char *attribute, int size,
                               ClipsErrorHandler handler)
{
  position = NULL;
  attribute_to_hash = copystring(attribute);
  if (type == ClipsString)
    hash_table = new wxHashTable(wxKEY_STRING, size);
  else if (type == ClipsInteger)
    hash_table = new wxHashTable(wxKEY_INTEGER, size);
  else hash_table = NULL;

  CurrentClipsErrorHandler = handler;
  noErrors = 0;
  inputFilename = NULL;
}

ClipsDatabase::~ClipsDatabase(void)
{
  ClearDatabase();
  if (hash_table)
    delete hash_table;
  if (attribute_to_hash)
    delete attribute_to_hash;
}

void ClipsDatabase::BeginFind(void)          // Initialise a search
{
  position = First();
}

ClipsExpr *ClipsDatabase::FindClause(long id)  // Find a term based on an integer id attribute
                                 // e.g. node(id=23, type=rectangle, ....).
{
  ClipsExpr *found = NULL;
  while (position && !found)
  {
    ClipsExpr *term = (ClipsExpr *)position->Data();
    
    if (term->Type() == ClipsList)
    {
      ClipsExpr *value = term->AttributeValue("id");
      if (value->Type() == ClipsInteger && value->IntegerValue() == id)
        found = term;
    }
    position = position->Next();
  }
  return found;
}

// Find on basis of attribute/value pairs, e.g. type=rectangle
ClipsExpr *ClipsDatabase::FindClause(char *word, char *val)
{
  ClipsExpr *found = NULL;
  while (position && !found)
  {
    ClipsExpr *term = (ClipsExpr *)position->Data();
    
    if (term->Type() == ClipsList)
    {
      ClipsExpr *value = term->AttributeValue(word);
      if ((value->Type() == ClipsWord && strcmp(value->WordValue(), val) == 0) ||
          (value->Type() == ClipsString && strcmp(value->StringValue(), val) == 0))
        found = term;
    }
    position = position->Next();
  }
  return found;
}

ClipsExpr *ClipsDatabase::FindClause(char *word, long val)
{
  ClipsExpr *found = NULL;
  while (position && !found)
  {
    ClipsExpr *term = (ClipsExpr *)position->Data();
    
    if (term->Type() == ClipsList)
    {
      ClipsExpr *value = term->AttributeValue(word);
      if ((value->Type() == ClipsInteger) && (value->IntegerValue() == val))
        found = term;
    }
    position = position->Next();
  }
  return found;
}

ClipsExpr *ClipsDatabase::FindClause(char *word, float val)
{
  ClipsExpr *found = NULL;
  while (position && !found)
  {
    ClipsExpr *term = (ClipsExpr *)position->Data();
    
    if (term->Type() == ClipsList)
    {
      ClipsExpr *value = term->AttributeValue(word);
      if ((value->Type() == ClipsReal) && (value->RealValue() == val))
        found = term;
    }
    position = position->Next();
  }
  return found;
}

ClipsExpr *ClipsDatabase::FindClauseByFunctor(char *functor)
{
  ClipsExpr *found = NULL;
  while (position && !found)
  {
    ClipsExpr *term = (ClipsExpr *)position->Data();
    
    if (term->Type() == ClipsList)
    {
      char *value = term->Functor();
      if (strcmp(value, functor) == 0)
        found = term;
    }
    position = position->Next();
  }
  return found;
}

// If hashing is on, must store in hash table too
void ClipsDatabase::Append(ClipsExpr *clause)
{
  wxList::Append((wxObject *)clause);
  if (hash_table)
  {
    char *functor = clause->Functor();
    ClipsExpr *expr = clause->AttributeValue(attribute_to_hash);
    if (expr)
    {
      long functor_key = hash_table->MakeKey(functor);
      long value_key = 0;
//      ClipsExpr *value_expr = (ClipsExpr *)(expr->ListValue()->Nth(2)->Data());
      if (expr && expr->Type() == ClipsString)
      {
        value_key = hash_table->MakeKey(expr->StringValue());
        hash_table->Put(functor_key + value_key, expr->StringValue(), (wxObject *)clause);
      }
      else if (expr && expr->Type() == ClipsInteger)
      {
        value_key = expr->IntegerValue();
        hash_table->Put(functor_key + value_key, expr->IntegerValue(), (wxObject *)clause);
      }

    }
  }
}

/*
// Clips compatibility
void ClipsDatabase::AddTemplate(ClipsTemplate *temp)
{
  templates.Append(temp);
}

// Find the template corresponding to the given name, and put a temp.
// pointer in the ClipsDatabase structure
ClipsTemplate *ClipsDatabase::FindTemplate(char *template_name)
{
  wxNode *node = templates.First();
  ClipsTemplate *temp = NULL;
  while (node && !temp)
  {
    ClipsTemplate *t = (ClipsTemplate *)node->Data();
    if (strcmp(template_name, t->name) == 0)
      temp = t;
    node = node->Next();
  }
  return temp;
}
*/

ClipsExpr *ClipsDatabase::HashFind(char *functor, long value)
{
  long key = hash_table->MakeKey(functor) + value;

  // The key alone isn't guaranteed to be unique:
  // must supply value too. Let's assume the value of the
  // id is going to be reasonably unique.
  return (ClipsExpr *)hash_table->Get(key, value);
}

ClipsExpr *ClipsDatabase::HashFind(char *functor, char *value)
{
  long key = hash_table->MakeKey(functor) + hash_table->MakeKey(value);
  return (ClipsExpr *)hash_table->Get(key, value);
}

void ClipsDatabase::ClearDatabase(void)
{
  noErrors = 0;
  wxNode *node = First();
  while (node)
  {
    ClipsExpr *expr = (ClipsExpr *)node->Data();
    delete expr;
    delete node;
    node = First();
  }
/*
  node = templates.First();
  while (node)
  {
    ClipsTemplate *temp = (ClipsTemplate *)node->Data();
    delete temp;
    delete node;
    node = First();
  }
*/
  if (hash_table)
    hash_table->Clear();
}

Bool ClipsDatabase::ReadClips(char *filename, wxNode *pos)
{
  PositionToInsert = pos;
  noErrors = 0;
  if (inputFilename) delete[] inputFilename;
  inputFilename = copystring(filename);
#ifdef wx_msw
  Unix2DosFilename(inputFilename);
#endif

  FILE *f = fopen(inputFilename, "r");
  if (f)
  {
    hyClipsDatabase = this;

    LexFromFile(f);
    yyparse();
    fclose(f);
    return (noErrors == 0);
  }
  else
  {
    return FALSE;
  }
}

Bool ClipsDatabase::ReadClipsFromString(char *buffer)
{
  noErrors = 0;
  if (buffer)
  {
    hyClipsDatabase = this;

    LexFromString(buffer);
    yyparse();
    return (noErrors == 0);
  }
  else
  {
    return FALSE;
  }
}


void ClipsDatabase::WriteClips(ostream& stream)
{
  noErrors = 0;
  wxNode *node = First();
  while (node)
  {
    ClipsExpr *expr = (ClipsExpr *)node->Data();
    expr->WriteClipsClause(stream);
    node = node->Next();
  }
}

void ClipsDatabase::WriteLisp(ostream& stream)
{
  noErrors = 0;
  wxNode *node = First();
  while (node)
  {
    ClipsExpr *expr = (ClipsExpr *)node->Data();
    expr->WriteLispExpr(stream);
    stream << "\n\n";
    node = node->Next();
  }
}

/*
// Write deftemplates and then facts
void ClipsDatabase::WriteAsClips(ostream& stream)
{
  noErrors = 0;
  wxNode *node = templates.First();
  while (node)
  {
    ClipsTemplate *temp = (ClipsTemplate *)(node->Data());
    temp->Write(stream);
    node = node->Next();
  }

  node = First();
  while (node)
  {
    ClipsExpr *expr = (ClipsExpr *)node->Data();
    expr->WriteClipsClause(stream);
    node = node->Next();
  }
}
*/

// Write templates, and filter out unknown slots when writing facts
/*
void ClipsDatabase::WriteClipsFiltering(ostream& stream)
{
  noErrors = 0;
  wxNode *node = templates.First();
  while (node)
  {
    ClipsTemplate *temp = (ClipsTemplate *)(node->Data());
    temp->Write(stream);
    node = node->Next();
  }

  node = First();
  while (node)
  {
    ClipsExpr *expr = (ClipsExpr *)node->Data();
    expr->WriteAsClipsClause(stream, TRUE, this);
    node = node->Next();
  }
}
*/

void add_expr(ClipsExpr * expr)
{
  if (PositionToInsert)
    hyClipsDatabase->Insert(PositionToInsert, (wxObject *)expr);
  else
    hyClipsDatabase->Append(expr);
}

/*
 * Called from parser
 *
 */

char *make_integer(char *str)
{
  ClipsExpr *x = new ClipsExpr(atol(str));

  return (char *)x;
}

char *make_real(char *str1, char *str2)
{
  char buf[50];

  sprintf(buf, "%s.%s", str1, str2);
  float f = (float)atof(buf);
  ClipsExpr *x = new ClipsExpr(f);

  return (char *)x;
}

char *make_word(char *str)
{
  ClipsExpr *x = new ClipsExpr(ClipsWord, str);
  return (char *)x;
}

char *make_string(char *str)
{
  char *s, *t;
  int len, i;

  str++;			/* skip leading quote */
  len = strlen(str) - 1;	/* ignore trailing quote */
    
  s = new char[len + 1];
    
  t = s;
  for(i=0; i<len; i++)
  {
    if (str[i] == '\\' && str[i+1] == '"')
    {
      *t++ = '"';
      i ++;
    }
    else if (str[i] == '\\' && str[i+1] == '\\')
    {
      *t++ = '\\';
      i ++;
    }
    else
      *t++ = str[i];
  }

  *t = '\0';

  ClipsExpr *x = new ClipsExpr(ClipsString, s, FALSE);
  return (char *)x;
}

char *cons(char * ccar, char * ccdr)
{
  ClipsExpr *car = (ClipsExpr *)ccar;
  ClipsExpr *cdr = (ClipsExpr *)ccdr;

  if (cdr == NULL)
  {
    cdr = new ClipsExpr(ClipsList);
  }
  if (car)
    cdr->Insert(car);
  return (char *)cdr;
}

void process_command(char * cexpr)
{
  ClipsExpr *expr = (ClipsExpr *)cexpr;
  add_expr(expr);
}

void syntax_error()
{
  if (CurrentClipsErrorHandler)
    (void)(*(CurrentClipsErrorHandler))(CLIPS_ERROR_SYNTAX, "syntax error");
  if (hyClipsDatabase) hyClipsDatabase->noErrors += 1;
}

#ifdef _WINDLL
char *__cdecl strdup(const char *s)
{
  int len = strlen(s);
  char *new_s = (char *)malloc(sizeof(char)*(len+1));
  strcpy(new_s, s);
  return new_s;
}
#endif

/*
 * Conversion data structures
 *
 */

ClipsConstruct::ClipsConstruct(char *name, int typ, ClipsExpr *val, ClipsType retType)
{
  constructValue = val;
  constructType = typ;
  if (name)
    constructName = copystring(name);
  else
    constructName = NULL;
  returnType = retType;
  parentClass = NULL;
  userType = NULL;
}

ClipsConstruct::~ClipsConstruct(void)
{
  if (constructName) delete[] constructName;
  if (parentClass) delete[] parentClass;
  if (userType) delete[] userType;
}

char *ClipsConstruct::GetCName(void)
{
  static char buf[200];
  char *name = GetName();
  if (name[0] == '?')
    name += 1;
  strcpy(buf, name);
  ClipsNameToC(buf);
  return buf;
}

NameSpace::NameSpace(void):wxHashTable(wxKEY_STRING)
{
}

NameSpace::~NameSpace(void)
{
  ClearConstructs();
}

ClipsConstruct *NameSpace::FindConstruct(char *name, char *className)
{
  char buf[200];
  strcpy(buf, name);
  if (className)
  {
    // Function is a member of a class
    strcat(buf, "*");
    strcat(buf, className);
  }
  return (ClipsConstruct *)Get(buf);
}

ClipsConstruct *NameSpace::AddConstruct(char *name, int typ, ClipsExpr *val, char *className)
{
  ClipsConstruct *constr = NULL;
  if (!(constr = FindConstruct(name, className)))
  {
    char buf[200];
    constr = new ClipsConstruct(name, typ, val);
    strcpy(buf, name);
    if (className)
    {
      // Function is a member of a class
      strcat(buf, "*");
      strcat(buf, className);
      constr->SetUserType(className);
    }
    Put(buf, constr);
  }
  return constr;
}

void NameSpace::ClearConstructs(void)
{
  DeleteContents(TRUE);
  Clear();
  DeleteContents(FALSE);
}

/*
 * Local variable list
 *
 */
 
ClipsConstruct *LocalVariables::FindConstruct(char *name)
{
  wxNode *node = Find(name);
  if (node)
    return (ClipsConstruct *)node->Data();
  else
    return NULL;
}

ClipsConstruct *LocalVariables::AddConstruct(char *name, int typ, ClipsExpr *val)
{
  ClipsConstruct *constr = NULL;
  if (!(constr = FindConstruct(name)))
  {
    constr = new ClipsConstruct(name, typ, val);
    Append(name, constr);
  }
  return constr;
}

/*
 * Function table
 *
 */
 
ClipsFunctionRecord *FunctionTable::FindFunction(char *name)
{
  return (ClipsFunctionRecord *)Get(name);
}

void FunctionTable::AddFunction(ClipsArg *returnType, char *clipsName, char *cName,
       ClipsArg *arg1, ClipsArg *arg2, ClipsArg *arg3, ClipsArg *arg4,
       ClipsArg *arg5, ClipsArg *arg6, ClipsArg *arg7, ClipsArg *arg8,
       ClipsArg *arg9, ClipsArg *arg10, ClipsArg *arg11, ClipsArg *arg12)
{
  ClipsFunctionRecord *record = new ClipsFunctionRecord(clipsName, cName);
  record->returnType = returnType->argType;
  delete returnType;
  
  if (arg1)
    record->argList.Append(arg1);
  if (arg2)
    record->argList.Append(arg2);
  if (arg3)
    record->argList.Append(arg3);
  if (arg4)
    record->argList.Append(arg4);
  if (arg5)
    record->argList.Append(arg5);
  if (arg6)
    record->argList.Append(arg6);
  if (arg7)
    record->argList.Append(arg7);
  if (arg8)
    record->argList.Append(arg8);
  if (arg9)
    record->argList.Append(arg9);
  if (arg10)
    record->argList.Append(arg10);
  if (arg11)
    record->argList.Append(arg11);
  if (arg12)
    record->argList.Append(arg12);
  Put(clipsName, record);
}

/*
 * Utilities
 *
 */

void ClipsNameToC(char *name)
{
  int len = strlen(name);
  for (int i = 0; i < len; i++)
  {
    int ch = name[i];
    if (ch == '-' || ch == '*')
      name[i] = '_';
  }
}

static Bool Clips2CEcho = TRUE;

void SetClipsEcho(Bool doEcho)
{
  Clips2CEcho = doEcho;
}

void ClipsWarning(char *msg)
{
  if (Clips2CEcho)
    OnClipsWarning(msg);
}

void ClipsError(char *msg)
{
  if (Clips2CEcho)
    OnClipsError(msg);
}

void ClipsInform(char *msg)
{
  if (Clips2CEcho)
    OnClipsInform(msg);
}

