/*
 * File:	clips2c.h
 * Purpose:	CLISP subset reading/manipulation
 * Author:	Julian Smart
 * Created:	1993
 * Updated:	
 * Copyright:	(c) 1993, AIAI, University of Edinburgh
 */

/* sccsid[] = "%W% %G%" */

#ifndef hy_clips2ch
#define hy_clips2ch

#include "clipslib.h"
#include "wx_utils.h"

// Error types
#define CLIPS_ERROR_GENERAL 1
#define CLIPS_ERROR_SYNTAX  2

// Conversion modes
#define MODE_PLAIN_CLIPS    1
#define MODE_WXCLIPS        2
#define MODE_HARDY          3

// Error handler function definition. If app returns TRUE,
// carry on processing.
typedef Bool (*ClipsErrorHandler) (int errorType, char *msg);

extern ClipsErrorHandler CurrentClipsErrorHandler;

extern "C" FILE *yyin;
extern "C" int yyparse(void);

class ClipsConstruct;
class ClipsDatabase: public wxList
{
  wxNode *position;              // Where we are in a search
  wxHashTable *hash_table;
  char *attribute_to_hash;
  char *inputFilename;
  wxList templates;
 public:
  int noErrors;

  ClipsDatabase(ClipsErrorHandler handler = 0);

  // Use hashing on both the functor, and the attribute of
  // specified type (ClipsString or ClipsInteger) and name.
  // So to find node 45
  // (i.e. match the clause node(id=45, ...))
  // it usually requires 1 look-up: the keys for functor and attribute
  // are added together.
  // Obviously if the attribute was missing in a clause, it would
  // fail to be found by this method, but could be retrieved by a
  // linear search using BeginFind and FindClauseByFunctor,
  // or just searching through the list as per usual.

  ClipsDatabase(ClipsType type, char *attribute, int size = 500,
                 ClipsErrorHandler handler = 0);

  ~ClipsDatabase(void);

  inline char *GetInputFile(void) { return inputFilename; }

  void BeginFind(void);          // Initialise a search
  ClipsExpr *FindClause(long id);  // Find a term based on an integer id attribute
                                 // e.g. node(id=23, type=rectangle, ....).

  // Find on basis of attribute/value pairs, e.g. type=rectangle
  // This doesn't use hashing; it's a linear search.
  ClipsExpr *FindClause(char *word, char *value);
  ClipsExpr *FindClause(char *word, long value);
  ClipsExpr *FindClause(char *word, float value);
  ClipsExpr *FindClauseByFunctor(char *functor);

  ClipsExpr *HashFind(char *functor, char *value);
  ClipsExpr *HashFind(char *functor, long value);

  void Append(ClipsExpr *expr);  // Does cleverer things if hashing is on
  void ClearDatabase(void);
  inline int GetErrorCount() { return noErrors; }
  Bool ReadClips(char *filename, wxNode *pos = NULL);
  Bool ReadClipsFromString(char *buffer);
  void WriteClips(ostream& stream);
  void WriteLisp(ostream& stream);

  /*
   * Code generation members
   *
   */
  Bool GenerateC(char *filename);
  
  // Phase 1: traverse CLIPS code setting up tables of variables and functions, etc.
  Bool Phase1(void);
  // Phase 2: resolve variable types and output code.
  Bool Phase2(ostream& stream, char *filename);

  Bool ProcessFunction(ClipsExpr *expr);
  Bool ProcessMessageHandler(ClipsExpr *expr);
  Bool ProcessClass(ClipsExpr *expr);
  Bool AddAccessorFunctions(ClipsConstruct *construct);
  Bool ProcessGlobal(ClipsExpr *expr);
  Bool ProcessStatement(ClipsExpr *expr);
  Bool ProcessLoad(ClipsExpr *expr, wxNode *pos);

  // Recursively process the expression in the given function,
  // returning an expression
  ClipsExpr *ProcessExpression(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent);
  ClipsExpr *ProcessIf(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent);
  ClipsExpr *ProcessWhile(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent);
  ClipsExpr *ProcessSwitch(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent);
  ClipsExpr *ProcessLoop(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent);
  ClipsExpr *ProcessProgn(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent);
  ClipsExpr *ProcessBind(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent);
  ClipsExpr *ProcessEq(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent);
  ClipsExpr *ProcessNeq(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent);
  ClipsExpr *ProcessUserFunction(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent);
  ClipsExpr *ProcessClipsFunction(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent);
  ClipsExpr *ProcessDeclare(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent);
  ClipsExpr *ProcessReturn(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent);
  ClipsExpr *ProcessStrcat(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent);
  ClipsExpr *ProcessMvAppend(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent);
  ClipsExpr *ProcessNth(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent);
  ClipsExpr *ProcessArithmetic(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent);
  ClipsExpr *ProcessComparison(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent);
  ClipsExpr *ProcessPredicate(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent);
  ClipsExpr *ProcessMember(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent);
  ClipsExpr *ProcessLength(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent);
  ClipsExpr *ProcessFormat(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent);
  ClipsExpr *ProcessPrintout(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent);

  // If this expression is a variable, resolve its type.
  void ResolveVariableType(ClipsConstruct *func, ClipsExpr *argExpr, ClipsType theType);

  Bool GenerateExpression(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent, ostream& stream);
  Bool GenerateClass(ClipsConstruct *func, ostream& stream);
  Bool GenerateIf(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent, ostream& stream);
  Bool GenerateWhile(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent, ostream& stream);
  Bool GenerateSwitch(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent, ostream& stream);
  Bool GenerateLoop(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent, ostream& stream);
  Bool GenerateProgn(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent, ostream& stream);
  Bool GenerateBreak(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent, ostream& stream);
  Bool GenerateBind(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent, ostream& stream);
  Bool GenerateEq(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent, ostream& stream);
  Bool GenerateNeq(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent, ostream& stream);
  Bool GenerateLogicalOp(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent, ostream& stream);
  Bool GenerateComparison(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent, ostream& stream);
  Bool GenerateArithmetic(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent, ostream& stream);
  Bool GenerateUserFunction(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent, ostream& stream);
  Bool GenerateClipsFunction(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent, ostream& stream);
  Bool GenerateReturn(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent, ostream& stream);
  Bool GenerateStrcat(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent, ostream& stream);
  Bool GenerateSubstring(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent, ostream& stream);
  Bool GenerateUpcase(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent, ostream& stream);
  Bool GenerateLowcase(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent, ostream& stream);
  Bool GeneratePredicate(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent, ostream& stream);
  Bool GenerateMvAppend(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent, ostream& stream);
  Bool GenerateNth(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent, ostream& stream);
  Bool GenerateMember(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent, ostream& stream);
  Bool GenerateInsert(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent, ostream& stream);
  Bool GenerateReplace(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent, ostream& stream);
  Bool GenerateFirstRest(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent, ostream& stream);

  Bool GenerateDeleteSubseq(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent, ostream& stream);
  Bool GenerateExplode(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent, ostream& stream);
  Bool GenerateImplode(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent, ostream& stream);
  Bool GenerateLength(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent, ostream& stream);
  Bool GenerateFormat(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent, ostream& stream);
  Bool GenerateOpen(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent, ostream& stream);
  Bool GenerateClose(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent, ostream& stream);
  Bool GeneratePrintout(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent, ostream& stream);
  Bool GenerateRead(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent, ostream& stream);
  Bool GenerateReadline(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent, ostream& stream);
  Bool GenerateMinMax(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent, ostream& stream);

  Bool GeneratePrototypes(ostream& stream);
  Bool GenerateFunction(ClipsConstruct *construct, ostream& stream);
  Bool GenerateMessageHandler(ClipsConstruct *construct, ostream& stream);
  Bool GenerateGlobal(ClipsConstruct *construct, ostream& stream);
  Bool GenerateStatement(ClipsConstruct *construct, ostream& stream);

  char *GenerateType(ClipsExpr *expr, ClipsType type);
};

// Temporary variable for communicating between read.cc and YACC/LEX
extern ClipsDatabase *hyClipsDatabase;

/*
 * Data structures for code generation
 *
 */

/*
 * Construct: may be function, global variable or local variable.
 *
 */
#define CONSTRUCT_TYPE_FUNCTION   1
#define CONSTRUCT_TYPE_GLOBAL     2
#define CONSTRUCT_TYPE_LOCAL      3
#define CONSTRUCT_TYPE_PARAMETER  4
#define CONSTRUCT_TYPE_STATEMENT  5
#define CONSTRUCT_TYPE_CLASS      6
#define CONSTRUCT_TYPE_INSTANCE   7
#define CONSTRUCT_TYPE_SLOT       8
#define CONSTRUCT_TYPE_HANDLER    9
#define CONSTRUCT_TYPE_UNKNOWN    20

class LocalVariables: public wxList
{
  public:
   inline LocalVariables(void):wxList(wxKEY_STRING)
   {
   }
   inline ~LocalVariables(void)
   {
     DeleteContents(TRUE);
   }
   ClipsConstruct *FindConstruct(char *name);
   ClipsConstruct *AddConstruct(char *name, int typ = CONSTRUCT_TYPE_UNKNOWN, ClipsExpr *val = NULL);
};

/*
 * Slot for class or instance
 *
 */
 
class ClipsConstruct: public wxObject
{
 public:
  int constructType;
  ClipsType returnType;
  ClipsExpr *constructValue;
  char *constructName;
  char *userType;   // Class name if this is a slot containing an object, or a function
  LocalVariables locals;
  wxList members; // Slots/handlers if class or instance
  char *parentClass; // isa class if class

  ClipsConstruct(char *name, int typ = CONSTRUCT_TYPE_UNKNOWN, ClipsExpr *val = NULL, ClipsType retType = ClipsUnknown);
  ~ClipsConstruct(void);

  inline char *GetName(void) { return constructName; }
  inline void SetName(char *s) { constructName = copystring(s); }
  char *GetCName(void);
  inline int GetType(void) { return constructType; }
  inline void SetType(int typ) { constructType = typ; }
  inline ClipsExpr *GetValue(void) { return constructValue; }
  inline void SetValue(ClipsExpr *val) { constructValue = val; }
  inline ClipsType GetReturnType(void) { return returnType; }
  inline void SetReturnType(ClipsType typ) { returnType = typ; }

  inline char *GetParentClass(void) { return parentClass; }
  inline void SetParentClass(char *s) { parentClass = copystring(s); }
  inline char *GetUserType(void) { return userType; }
  inline void SetUserType(char *s) { userType = copystring(s); }
};

class NameSpace: public wxHashTable
{
 public:
  NameSpace(void);
  ~NameSpace(void);

  ClipsConstruct *FindConstruct(char *name, char *className = NULL);
  ClipsConstruct *AddConstruct(char *name, int typ = CONSTRUCT_TYPE_UNKNOWN, ClipsExpr *val = NULL,
       char *className = NULL);
  void ClearConstructs(void);
};

extern NameSpace TheNameSpace;

/*
 * Mapping from CLIPS functions to C functions
 *
 */

class ClipsArg: public wxObject
{
 public:
  ClipsType argType;
  Bool isOptional;
  inline ClipsArg(ClipsType typ, Bool isOpt = FALSE)
  {
    argType = typ;
    isOptional = isOpt;
  }
};

#define cString (new ClipsArg(ClipsString))
#define cStringOpt (new ClipsArg(ClipsString, TRUE))

#define cStringClass (new ClipsArg(ClipsStringClass))
#define cStringClassOpt (new ClipsArg(ClipsStringClass, TRUE))

#define cInteger (new ClipsArg(ClipsInteger))
#define cIntegerOpt (new ClipsArg(ClipsInteger, TRUE))

#define cReal (new ClipsArg(ClipsReal))
#define cRealOpt (new ClipsArg(ClipsReal, TRUE))

#define cList (new ClipsArg(ClipsList))
#define cListOpt (new ClipsArg(ClipsList, TRUE))

#define cFunction (new ClipsArg(ClipsFunction))
#define cFunctionOpt (new ClipsArg(ClipsFunction, TRUE))

#define cAny (new ClipsArg(ClipsAny))
#define cAnyOpt (new ClipsArg(ClipsAny, TRUE))

#define cUnknown (new ClipsArg(ClipsUnknown))
#define cUnknownOpt (new ClipsArg(ClipsUnknown, TRUE))

#define cVoid (new ClipsArg(ClipsNull))

class ClipsFunctionRecord: public wxObject
{
 public:
  char *clipsFunctionName;
  char *cFunctionName;
  ClipsType returnType;
  wxList argList;
 
  ClipsFunctionRecord(char *clipsName, char *cName)
  {
    clipsFunctionName = copystring(clipsName);
    cFunctionName = copystring(cName);
    returnType = ClipsUnknown;
  }
  ~ClipsFunctionRecord(void)
  {
    argList.DeleteContents(TRUE);
    delete[] clipsFunctionName;
    delete[] cFunctionName;
  }
  inline char *GetClipsName(void) { return clipsFunctionName; }
  inline char *GetCName(void) { return cFunctionName; }
  inline ClipsType GetReturnType (void) { return returnType; }
};

class FunctionTable: public wxHashTable
{
 public:
  inline FunctionTable(void):
   wxHashTable(wxKEY_STRING)
  {
  }
  inline ~FunctionTable(void)
  {
    DeleteContents(TRUE);
  }

  // Give CLIPS name
  ClipsFunctionRecord *FindFunction(char *name);

  void AddFunction(ClipsArg *returnType, char *clipsName, char *cName,
       ClipsArg *arg1 = NULL, ClipsArg *arg2 = NULL, ClipsArg *arg3 = NULL, ClipsArg *arg4 = NULL,
       ClipsArg *arg5 = NULL, ClipsArg *arg6 = NULL, ClipsArg *arg7 = NULL, ClipsArg *arg8 = NULL,
       ClipsArg *arg9 = NULL, ClipsArg *arg10 = NULL, ClipsArg *arg11 = NULL, ClipsArg *arg12 = NULL);
};

extern FunctionTable TheFunctionTable;
extern wxList ConstructList;
void ClipsNameToC(char *name);
void InitializeClips2C(int theMode);

#define AddClipsFunction TheFunctionTable.AddFunction

void SetClipsEcho(Bool doEcho);

void ClipsWarning(char *msg);
void ClipsError(char *msg);
void ClipsInform(char *msg);

/*
 * Define in client program
 *
 */
 
void OnClipsWarning(char *msg);
void OnClipsError(char *msg);
void OnClipsInform(char *msg);

void SetGenerationMode(int mode);
int GetGenerationMode(void);

#endif // hy_clips2ch
