/*
 * File:	clipslib.cc
 * Purpose:	Run-time support for CLIPS2C generated code
 * Author:	Julian Smart
 * Created:	1993
 * Updated:	
 * Copyright:	(c) 1993, AIAI, University of Edinburgh
 */

static const char sccsid[] = "%W% %G%";

#include <stdarg.h>
#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>
#include <iostream.h>
#ifdef SYSV
#include <varargs.h>
#endif

#include <time.h>
#include <sys/types.h>

#if defined(SVR4) || defined(SYSV) || defined(__sgi)
#include <sys/time.h>
#else
#include <sys/timeb.h>
#endif

#ifdef sun
// At least on Sun, ftime is undeclared.
// Need to be verified on other platforms.
extern "C" int ftime(struct timeb *tp);
#ifdef SVR4
// dito for gettimeofday on Solaris 2.x.
extern "C" int gettimeofday(struct timeval *tp);
#endif
#endif

#include "clipslib.h"

extern long NewId(void);

ClipsStreamList TheClipsStreamList;
extern void ClipsErrorFunction(char *s);

// Copy constructor
ClipsExpr::ClipsExpr(ClipsExpr& expr)
{
  next = NULL;
  last = NULL;
  type = expr.type;
  returnType = expr.returnType;
  client_data = NULL;
  switch (type)
  {
    case ClipsInteger:
      value.integer = expr.value.integer;
      break;
    case ClipsReal:
      value.real = expr.value.real;
      break;
    case ClipsString:
      if (expr.value.string)
        value.string = copystring(expr.value.string);
      else
        value.string = NULL;
      break;
    case ClipsUser:
      if (expr.value.string)
        value.string = copystring(expr.value.string);
      else
        value.string = NULL;
      break;
    case ClipsStringClass:
    {
      char *data = expr.value.stringClass->GetData();
      value.stringClass = new wxString;
      if (data)
        value.stringClass->SetData(copystring(data));
      else
        value.stringClass->SetData(NULL);
      break;
    }
    case ClipsWord:
      if (expr.value.word)
        value.word = copystring(expr.value.word);
      else
        value.word = NULL;
      break;
    case ClipsList:
    {
      next = NULL;
      last = NULL;
      value.first = NULL;
      ClipsExpr *currExpr = expr.value.first;
      while (currExpr)
      {
        ClipsExpr *expr2 = new ClipsExpr(*currExpr);
        Append(expr2);
        currExpr = currExpr->GetNext();
      }
      break;
    }
   case ClipsUnknown:
   case ClipsNull: 
    break;
  }
}

ClipsExpr::ClipsExpr(char *functor)
{
  type = ClipsList;
  returnType = ClipsUnknown;
  next = NULL;
  last = NULL;
  value.first = NULL;

  ClipsExpr *pfunctor = new ClipsExpr(ClipsWord, functor);
  Append(pfunctor);
  client_data = NULL;
}

ClipsExpr::ClipsExpr(ClipsType the_type, char *word_or_string, Bool allocate)
{
  type = the_type;
  returnType = ClipsUnknown;
  next = NULL;
  last = NULL;
  value.first = NULL;
  client_data = NULL;

  switch (the_type)
  {
   case ClipsWord:
    value.word = allocate ? copystring(word_or_string) : word_or_string;
    returnType = ClipsWord;
    break;
   case ClipsString:
    value.string = allocate ? copystring(word_or_string) : word_or_string;
    returnType = ClipsString;
    break;
   case ClipsUser:
    value.string = allocate ? copystring(word_or_string) : word_or_string;
    returnType = ClipsString;
    break;
   case ClipsList:
    last = NULL;
    value.first = NULL;
    break;
   case ClipsReal:
   {
     returnType = ClipsWord;
     value.real = 0.0;
     break;
   }
   case ClipsInteger:
   {
     returnType = ClipsInteger;
     value.integer = 0;
     break;
   }
   case ClipsNull: 
   case ClipsUnknown: 
     break;
  }
}

ClipsExpr::ClipsExpr(long the_integer)
{
  type = ClipsInteger;
  returnType = ClipsInteger;
  value.integer = the_integer;
  client_data = NULL;
  next = NULL;
}

ClipsExpr::ClipsExpr(float the_real)
{
  type = ClipsReal;
  returnType = ClipsReal;
  value.real = the_real;
  client_data = NULL;
  next = NULL;
}

ClipsExpr::ClipsExpr(wxString& string)
{
  type = ClipsStringClass;
  returnType = ClipsStringClass;
  value.stringClass = new wxString(string);
  client_data = NULL;
  next = NULL;
}

ClipsExpr::ClipsExpr(wxList *the_list)
{
  type = ClipsList;
  returnType = ClipsUnknown;
  client_data = NULL;
  last = NULL;
  value.first = NULL;

  ClipsExpr *listExpr = new ClipsExpr(ClipsList);

  wxNode *node = the_list->First();
  while (node)
  {
    ClipsExpr *expr = (ClipsExpr *)node->Data();
    listExpr->Append(expr);
    node = node->Next();
  }
  Append(listExpr);

  delete the_list;
}

ClipsExpr::~ClipsExpr(void)
{
  switch (type)
  {
    case ClipsInteger:
    case ClipsReal:
    {
     break;
    }
   case ClipsString:
   {
     delete[] value.string;
     break;
   }
   case ClipsStringClass:
   {
     delete value.stringClass;
     break;
   }
   case ClipsWord:
   {
     delete value.word;
     break;
   }
   case ClipsList:
   {
     ClipsExpr *expr = value.first;
     while (expr)
     {
       ClipsExpr *expr1 = expr->GetNext();

       delete expr;
       expr = expr1;
     }
     break;
   }
   case ClipsUnknown:
   case ClipsNull: break;
  }
}

void ClipsExpr::Append(ClipsExpr *expr)
{
  if (!value.first)
    value.first = expr;

  if (last)
    last->next = expr;
  last = expr;
}

void ClipsExpr::Insert(ClipsExpr *expr)
{
  expr->next = value.first;
  value.first = expr;

  if (!last)
    last = expr;
}

void ClipsExpr::operator = (ClipsExpr& expr)
{
  switch (expr.type)
  {
    case ClipsInteger:
      value.integer = expr.value.integer;
      break;
    case ClipsReal:
      value.real = expr.value.real;
      break;
    case ClipsString:
    {
      if (type == ClipsString && value.string)
        delete[] value.string;

      if (expr.value.string)
        value.string = copystring(expr.value.string);
      else
        value.string = NULL;
      break;
    }
    case ClipsStringClass:
    {
      value.stringClass = expr.value.stringClass;
      break;
    }
    case ClipsWord:
    {
      if (type == ClipsWord && value.word)
        delete[] value.word;

      if (expr.value.word)
        value.word = copystring(expr.value.word);
      else
        value.word = NULL;
      break;
    }
    case ClipsList:
    {
      if (type == ClipsList)
      {
        // Clear list
        ClipsExpr *currExpr = value.first;
        while (currExpr)
        {
          ClipsExpr *expr1 = currExpr->GetNext();

          delete currExpr;
          currExpr = expr1;
        }
        value.first = NULL;
        last = NULL;
      }
        
      ClipsExpr *currExpr = expr.value.first;
      while (currExpr)
      {
        ClipsExpr *expr2 = new ClipsExpr(*currExpr);
        Append(expr2);
        currExpr = currExpr->GetNext();
      }
      break;
    }
    case ClipsUnknown:
    case ClipsNull: 
      break;
  }
  type = expr.type;
}

float ClipsExpr::NthReal(int arg)
{
  ClipsExpr *valExpr = Nth(arg - 1);
  if (valExpr)
  {
    switch (valExpr->Type())
    {
      case ClipsReal:
      {
        return valExpr->RealValue();
        break;
      }
      case ClipsInteger:
      {
        return (float)valExpr->IntegerValue();
        break;
      }
      case ClipsString:
      case ClipsStringClass:
      case ClipsWord:
      case ClipsList:
      case ClipsNull:
      case ClipsUnknown:
      default:
      {
        return 0.0;
        break;
      }
    }
  }
  return 0.0;
}

long ClipsExpr::NthInteger(int arg)
{
  ClipsExpr *valExpr = Nth(arg - 1);
  if (valExpr)
  {
    switch (valExpr->Type())
    {
      case ClipsReal:
      {
        return (long)valExpr->RealValue();
        break;
      }
      case ClipsInteger:
      {
        return valExpr->IntegerValue();
        break;
      }
      case ClipsString:
      case ClipsStringClass:
      case ClipsWord:
      case ClipsList:
      case ClipsNull:
      case ClipsUnknown:
      default:
      {
        return 0;
        break;
      }
    }
  }
  return 0;
}

wxString ClipsExpr::NthString(int arg)
{
  ClipsExpr *valExpr = Nth(arg - 1);
  if (valExpr)
  {
    switch (valExpr->Type())
    {
      case ClipsString:
      {
        return wxString(valExpr->StringValue());
        break;
      }
      case ClipsWord:
      {
        return wxString(valExpr->WordValue());
        break;
      }
      case ClipsStringClass:
      {
        return valExpr->StringClassValue();
        break;
      }
      case ClipsInteger:
      case ClipsReal:
      case ClipsList:
      case ClipsNull:
      case ClipsUnknown:
      default:
      {
        return wxString((char *)NULL);
        break;
      }
    }
  }
  return wxString((char *)NULL);
}

ClipsExpr ClipsExpr::NthList(int arg)
{
  return ClipsExpr(ClipsUnknown);
}

int ClipsExpr::Number(void)
{
  if (type == ClipsList)
  {
    int n = 0;
    ClipsExpr *child = GetFirst();
    while (child)
    {
      child = child->GetNext();
      n ++;
    }
    return n;
  }
  else return 0;
}

int ClipsExpr::Member(float arg)
{
  int n = 0;
  if (Type() != ClipsList)
    return 0;
    
  ClipsExpr *expr = GetFirst();
  while (expr)
  {
    n ++;
    if ((expr->Type() == ClipsReal) && (expr->RealValue() == arg))
      return n;
    else if ((expr->Type() == ClipsInteger) && (expr->IntegerValue() == arg))
      return n;
    expr = expr->GetNext();
  }
  return 0;
}

int ClipsExpr::Member(long arg)
{
  int n = 0;
  if (Type() != ClipsList)
    return 0;
    
  ClipsExpr *expr = GetFirst();
  while (expr)
  {
    n ++;
    if ((expr->Type() == ClipsReal) && (expr->RealValue() == arg))
      return n;
    else if ((expr->Type() == ClipsInteger) && (expr->IntegerValue() == arg))
      return n;
    expr = expr->GetNext();
  }
  return 0;
}

int ClipsExpr::Member(wxString& arg)
{
  int n = 0;
  if (Type() != ClipsList)
    return 0;
    
  ClipsExpr *expr = GetFirst();
  while (expr)
  {
    n ++;
    if ((expr->Type() == ClipsStringClass) && (expr->StringClassValue() == arg))
      return n;
    else if ((expr->Type() == ClipsString || expr->Type() == ClipsWord) &&
              (strcmp(expr->StringValue(), arg.GetData()) == 0))
      return n;
    expr = expr->GetNext();
  }
  return n;
}

ClipsExpr ClipsExpr::DeleteRange(int from, int to)
{
  ClipsExpr newList;
  int n = Number();
  int i;
  for (i = 1; i <= from; i++)
  {
    ClipsExpr *oldExpr = Nth(i-1);
    if (oldExpr)
      newList.Append(new ClipsExpr(*oldExpr));
  }
  for (i = to; i <= n; i++)
  {
    ClipsExpr *oldExpr = Nth(i-1);
    if (oldExpr)
      newList.Append(new ClipsExpr(*oldExpr));
  }
  return newList;
}

wxString ClipsExpr::Implode(void)
{
  static char buf[500];
  buf[0] = 0;
  static char buf1[50];
  ClipsExpr *expr = GetFirst();
  while (expr)
  {
    switch (expr->Type())
    {
      case ClipsReal:
      {
        sprintf(buf1, "%.4f", expr->RealValue());
        strcat(buf, buf1);
        break;
      }
      case ClipsInteger:
      {
        sprintf(buf1, "%l", expr->IntegerValue());
        strcat(buf, buf1);
        break;
      }
      case ClipsWord:
      case ClipsUser:
      case ClipsString:
      {
        strcat(buf, expr->StringValue());
        break;
      }
      case ClipsNull:
      {
        strcat(buf, "nil");
        break;
      }
      case ClipsUnknown:
      {
        strcat(buf, "<unknown>");
        break;
      }
      default:
        break;
    }
    expr = expr->GetNext();
    if (expr)
      strcat(buf, " ");
  }
  return wxString(buf);
}

ClipsExpr ClipsExpr::Subseq(int from, int to)
{
  ClipsExpr l;
  for (int i = from; i <= to; i++)
  {
    ClipsExpr *expr = Nth(i-1);
    if (expr)
      l.Append(new ClipsExpr(*expr));
  }
  return l;
}

ClipsExpr ClipsExpr::Replace(int from, int to, ClipsExpr& expr)
{
  ClipsExpr l;
  int i;
  for (i = 1; i <= from; i++)
  {
    ClipsExpr *expr1 = Nth(i-1);
    if (expr1)
      l.Append(new ClipsExpr(*expr1));
  }
  for (i = 0; i < expr.Number(); i++)
  {
    ClipsExpr *expr1 = expr.Nth(i);
    if (expr1)
      l.Append(new ClipsExpr(*expr1));
  }
  for (i = to; i <= Number(); i++)
  {
    ClipsExpr *expr1 = Nth(i-1);
    if (expr1)
      l.Append(new ClipsExpr(*expr1));
  }
  
  return l;
}

ClipsExpr ClipsExpr::Replace(int from, int to, float arg)
{
  ClipsExpr l;
  int i;
  for (i = 1; i <= from; i++)
  {
    ClipsExpr *expr = Nth(i-1);
    if (expr)
      l.Append(new ClipsExpr(*expr));
  }
  l.Append(new ClipsExpr(arg));
  for (i = to; i <= Number(); i++)
  {
    ClipsExpr *expr = Nth(i-1);
    if (expr)
      l.Append(new ClipsExpr(*expr));
  }
  return l;
}

ClipsExpr ClipsExpr::Replace(int from, int to, long arg)
{
  ClipsExpr l;
  int i;
  for (i = 1; i <= from; i++)
  {
    ClipsExpr *expr = Nth(i-1);
    if (expr)
      l.Append(new ClipsExpr(*expr));
  }
  l.Append(new ClipsExpr(arg));
  for (i = to; i <= Number(); i++)
  {
    ClipsExpr *expr = Nth(i-1);
    if (expr)
      l.Append(new ClipsExpr(*expr));
  }
  return l;
}

ClipsExpr ClipsExpr::Replace(int from, int to, wxString& arg)
{
  ClipsExpr l;
  int i;
  for (i = 1; i <= from; i++)
  {
    ClipsExpr *expr = Nth(i-1);
    if (expr)
      l.Append(new ClipsExpr(*expr));
  }
  l.Append(new ClipsExpr(arg));
  for (i = to; i <= Number(); i++)
  {
    ClipsExpr *expr = Nth(i-1);
    if (expr)
      l.Append(new ClipsExpr(*expr));
  }
  return l;
}

ClipsExpr ClipsExpr::Insert(int pos, ClipsExpr& expr)
{
  ClipsExpr l;
  int i;
  for (i = 1; i < pos; i++)
  {
    ClipsExpr *expr1 = Nth(i-1);
    if (expr1)
      l.Append(new ClipsExpr(*expr1));
  }
  for (i = 0; i < expr.Number(); i++)
  {
    ClipsExpr *expr1 = expr.Nth(i);
    if (expr1)
      l.Append(new ClipsExpr(*expr1));
  }
  for (i = pos; i <= Number(); i++)
  {
    ClipsExpr *expr1 = Nth(i-1);
    if (expr1)
      l.Append(new ClipsExpr(*expr1));
  }
  return l;
}

ClipsExpr ClipsExpr::Insert(int pos, float arg)
{
  ClipsExpr l;
  int i;
  for (i = 1; i < pos; i++)
  {
    ClipsExpr *expr = Nth(i-1);
    if (expr)
      l.Append(new ClipsExpr(*expr));
  }
  l.Append(new ClipsExpr(arg));
  for (i = pos; i <= Number(); i++)
  {
    ClipsExpr *expr = Nth(i-1);
    if (expr)
      l.Append(new ClipsExpr(*expr));
  }
  return l;
}

ClipsExpr ClipsExpr::Insert(int pos, long arg)
{
  ClipsExpr l;
  int i;
  for (i = 1; i < pos; i++)
  {
    ClipsExpr *expr = Nth(i-1);
    if (expr)
      l.Append(new ClipsExpr(*expr));
  }
  l.Append(new ClipsExpr(arg));
  for (i = pos; i <= Number(); i++)
  {
    ClipsExpr *expr = Nth(i-1);
    if (expr)
      l.Append(new ClipsExpr(*expr));
  }
  return l;
}

ClipsExpr ClipsExpr::Insert(int pos, wxString& arg)
{
  ClipsExpr l;
  int i;
  for (i = 1; i < pos; i++)
  {
    ClipsExpr *expr = Nth(i-1);
    if (expr)
      l.Append(new ClipsExpr(*expr));
  }
  l.Append(new ClipsExpr(arg));
  for (i = pos; i <= Number(); i++)
  {
    ClipsExpr *expr = Nth(i-1);
    if (expr)
      l.Append(new ClipsExpr(*expr));
  }
  return l;
}

ClipsExpr ClipsExpr::FirstList(void)
{
  ClipsExpr l;
  ClipsExpr *fExpr = GetFirst();
  if (fExpr)
    l.Append(new ClipsExpr(*fExpr));
  return l;
}

ClipsExpr ClipsExpr::Rest(void)
{
  ClipsExpr l;
  ClipsExpr *fExpr = GetFirst();
  if (fExpr)
  {
    fExpr = fExpr->GetNext();
    while (fExpr)
    {
      l.Append(new ClipsExpr(*fExpr));
      fExpr = fExpr->GetNext();
    }
  }
  return l;
}

// Get the ClipsExpr (containing (= ClipsExpr Value) form) for the given word
//  or string, assuming that we have Attribute=Value, ...
ClipsExpr *ClipsExpr::GetAttributeValueNode(char *word)  // Use only for a clause or list
{
  if (type != ClipsList)
    return NULL;

  ClipsExpr *expr = value.first;
  while (expr)
  {
    if (expr->type == ClipsList)
    {
      ClipsExpr *firstNode = expr->value.first;
      if ((firstNode->type == ClipsWord) && (firstNode->value.word[0] == '='))
      {
        ClipsExpr *secondNode = firstNode->GetNext();
        if ((secondNode->type == ClipsWord) && 
            (strcmp(word, secondNode->value.word) == 0))
        {
          return expr;
        }
      }
    }
    expr = expr->next;
  }
  return NULL;
}

// Get the value (in ClipsExpr form) for the given word or string, assuming
// that we have Attribute=Value, ...
ClipsExpr *ClipsExpr::AttributeValue(char *word)  // Use only for a clause or list
{
  if (type != ClipsList)
    return NULL;

  ClipsExpr *attExpr = GetAttributeValueNode(word);
  if (attExpr && attExpr->value.first && attExpr->value.first->next)
    return attExpr->value.first->next->next;
  else return NULL;
}

char *ClipsExpr::Functor(void)  // Use only for a clause
{
  if ((type != ClipsList) || !value.first)
    return NULL;

  if (value.first->type == ClipsWord)
    return value.first->value.word;
  else
    return NULL;
}

Bool ClipsExpr::IsFunctor(char *f)  // Use only for a clause
{
  if ((type != ClipsList) || !value.first)
    return FALSE;

  return (value.first->type == ClipsWord && 
          (strcmp(f, value.first->value.word) == 0));
}

// Return nth argument of a clause (starting from 1)
ClipsExpr *ClipsExpr::Arg(ClipsType type, int arg)
{
  ClipsExpr *expr = value.first;
  for (int i = 1; i < arg; i++)
    if (expr)
      expr = expr->next;

  if (expr && (expr->type == type))
    return expr;
  else
    return NULL;
}

// Return nth argument of a list expression (starting from zero)
ClipsExpr *ClipsExpr::Nth(int arg)
{
  if (type != ClipsList)
    return NULL;

  ClipsExpr *expr = value.first;
  for (int i = 0; i < arg; i++)
    if (expr)
      expr = expr->next;
    else return NULL;

  if (expr)
    return expr;
  else
    return NULL;
}

void ClipsExpr::DeleteAttributeValue(char *attribute)
{
  if (type != ClipsList)
    return;

  ClipsExpr *expr = value.first;
  ClipsExpr *lastExpr = this;
  while (expr)
  {
    if (expr->type == ClipsList)
    {
      ClipsExpr *firstNode = expr->value.first;
      if ((firstNode->type == ClipsWord) && (firstNode->value.word[0] == '='))
      {
        ClipsExpr *secondNode = firstNode->next;
        if ((secondNode->type == ClipsWord) && 
            (strcmp(attribute, secondNode->value.word) == 0))
        {
          ClipsExpr *nextExpr = expr->next;
          delete expr;

          lastExpr->next = nextExpr;

          if (last == expr)
            last = lastExpr;

          return;
        }
      }
    }
    lastExpr = expr;
    expr = expr->next;
  }
  return;
}

void ClipsExpr::AddAttributeValue(char *attribute, ClipsExpr *val)
{
  if (type != ClipsList)
  {
//    cout << "Error! tried to add an attribute-value pair to a nonlist Clips expression\n";
    return;
  }
  // Warning - existing code may assume that any existing value
  // is deleted first. For efficiency, we leave this to the application.
//  DeleteAttributeValue(attribute);

  ClipsExpr *patt = new ClipsExpr(ClipsWord, attribute);
  ClipsExpr *pequals = new ClipsExpr(ClipsWord, "=");

  ClipsExpr *listExpr = new ClipsExpr(ClipsList);

  listExpr->Append(pequals);
  listExpr->Append(patt);
  listExpr->Append(val);

  Append(listExpr);
}

void ClipsExpr::AddAttributeValue(char *attribute, long val)
{
  if (type != ClipsList)
  {
//    cout << "Error! tried to add an attribute-value pair to a nonlist Clips expression\n";
    return;
  }
  // Warning - existing code may assume that any existing value
  // is deleted first. For efficiency, we leave this to the application.
//  DeleteAttributeValue(attribute);

  ClipsExpr *patt = new ClipsExpr(ClipsWord, attribute);
  ClipsExpr *pval = new ClipsExpr(val);
  ClipsExpr *pequals = new ClipsExpr(ClipsWord, "=");

  ClipsExpr *listExpr = new ClipsExpr(ClipsList);

  listExpr->Append(pequals);
  listExpr->Append(patt);
  listExpr->Append(pval);

  Append(listExpr);
}

void ClipsExpr::AddAttributeValue(char *attribute, float val)
{
  if (type != ClipsList)
  {
//    cout << "Error! tried to add an attribute-value pair to a nonlist Clips expression\n";
    return;
  }

//  DeleteAttributeValue(attribute);
  ClipsExpr *patt = new ClipsExpr(ClipsWord, attribute);
  ClipsExpr *pval = new ClipsExpr(val);
  ClipsExpr *pequals = new ClipsExpr(ClipsWord, "=");

  ClipsExpr *listExpr = new ClipsExpr(ClipsList);

  listExpr->Append(pequals);
  listExpr->Append(patt);
  listExpr->Append(pval);

  Append(listExpr);
}

void ClipsExpr::AddAttributeValueString(char *attribute, char *val)
{
  if (type != ClipsList)
  {
//    cout << "Error! tried to add an attribute-value pair to a nonlist Clips expression\n";
    return;
  }
  if (!val)
    return;

//  DeleteAttributeValue(attribute);

  ClipsExpr *patt = new ClipsExpr(ClipsWord, attribute);
  ClipsExpr *pval = new ClipsExpr(ClipsString, val);
  ClipsExpr *pequals = new ClipsExpr(ClipsWord, "=");

  ClipsExpr *listExpr = new ClipsExpr(ClipsList);

  listExpr->Append(pequals);
  listExpr->Append(patt);
  listExpr->Append(pval);

  Append(listExpr);
}

void ClipsExpr::AddAttributeValueWord(char *attribute, char *val)
{
  if (type != ClipsList)
  {
//    cout << "Error! tried to add an attribute-value pair to a nonlist Clips expression\n";
    return;
  }
  if (!val)
    return;

//  DeleteAttributeValue(attribute);

  ClipsExpr *patt = new ClipsExpr(ClipsWord, attribute);
  ClipsExpr *pval = new ClipsExpr(ClipsWord, val);
  ClipsExpr *pequals = new ClipsExpr(ClipsWord, "=");

  ClipsExpr *listExpr = new ClipsExpr(ClipsList);

  listExpr->Append(pequals);
  listExpr->Append(patt);
  listExpr->Append(pval);

  Append(listExpr);
}

void ClipsExpr::AddAttributeValue(char *attribute, wxList *val)
{
  if (type != ClipsList)
  {
//    cout << "Error! tried to add an attribute-value pair to a nonlist Clips expression\n";
    return;
  }
  if (!val)
    return;

//  DeleteAttributeValue(attribute);

  ClipsExpr *patt = new ClipsExpr(ClipsWord, attribute);
  ClipsExpr *pval = new ClipsExpr(val);
  ClipsExpr *pequals = new ClipsExpr(ClipsWord, "=");

  ClipsExpr *listExpr = new ClipsExpr(ClipsList);

  listExpr->Append(pequals);
  listExpr->Append(patt);
  listExpr->Append(pval);

  Append(listExpr);
}

void ClipsExpr::AddAttributeValueStringList(char *attribute, wxList *string_list)
{
  if (type != ClipsList)
  {
//    cout << "Error! tried to add an attribute-value pair to a nonlist Clips expression\n";
    return;
  }
  if (!string_list)
    return;

//  DeleteAttributeValue(attribute);

  // First make a list of ClipsExpr strings
  ClipsExpr *listExpr = new ClipsExpr(ClipsList);
  wxNode *node = string_list->First();
  while (node)
  {
    char *string = (char *)node->Data();
    ClipsExpr *expr = new ClipsExpr(ClipsString, string);
    listExpr->Append(expr);
    node = node->Next();
  }

  // Now make an (=, Att, Value) triple
  ClipsExpr *patt = new ClipsExpr(ClipsWord, attribute);
  ClipsExpr *pequals = new ClipsExpr(ClipsWord, "=");

  ClipsExpr *listExpr2 = new ClipsExpr(ClipsList);

  listExpr2->Append(pequals);
  listExpr2->Append(patt);
  listExpr2->Append(listExpr);

  Append(listExpr2);
}

void ClipsExpr::AssignAttributeValue(char *att, int *var)
{
  ClipsExpr *expr = AttributeValue(att);

  if (expr && (expr->Type() == ClipsInteger || expr->Type() == ClipsReal))
    *var = (int)(expr->IntegerValue());
}

void ClipsExpr::AssignAttributeValue(char *att, long *var)
{
  ClipsExpr *expr = AttributeValue(att);

  if (expr && (expr->Type() == ClipsInteger || expr->Type() == ClipsReal))
    *var = expr->IntegerValue();
}

void ClipsExpr::AssignAttributeValue(char *att, float *var)
{
  ClipsExpr *expr = AttributeValue(att);
  if (expr && (expr->Type() == ClipsInteger || expr->Type() == ClipsReal))
    *var = expr->RealValue();
}

void ClipsExpr::AssignAttributeValue(char *att, char **var)  // Word OR string -> string
{
  ClipsExpr *expr = AttributeValue(att);
  if (expr && expr->Type() == ClipsWord)
    *var = copystring(expr->WordValue());
  else if (expr && expr->Type() == ClipsString)
    *var = copystring(expr->StringValue());
}

void ClipsExpr::AssignAttributeValue(char *att, ClipsExpr **var)
{
  ClipsExpr *expr = AttributeValue(att);
  if (expr)
    *var = expr;
}

void ClipsExpr::AssignAttributeValueStringList(char *att, wxList *var)
{
  ClipsExpr *expr = AttributeValue(att);
  if (expr && expr->Type() == ClipsList)
  {
    ClipsExpr *string_expr = expr->value.first;
    while (string_expr)
    {
      if (string_expr->Type() == ClipsString)
        var->Append((wxObject *)copystring(string_expr->StringValue()));

      string_expr = string_expr->next;
    }
  }
}

void ClipsExpr::WriteClipsClause(ostream& stream)  // Write this expression as a top-level clause
{
  if (type != ClipsList)
    return;

  if (!value.first)
    stream << "()";
  else
  {
    // Write all constructs
    ClipsExpr *expr = value.first;
    if (expr->Type() == ClipsWord)
    {
      char *val = expr->WordValue();
      if (strcmp(val, "deffunction") == 0)
      {
        stream << "(deffunction ";
        expr = expr->next;
        expr->WriteClipsExpr(stream);
        stream << " ";
        expr = expr->next;
        while (expr)
        {
          expr->WriteClipsExpr(stream);
          expr = expr->next;
          if (expr) stream << "\n";
	}
        stream << ")\n\n";
      }
      else
      {
        stream << "(";
        while (expr)
        {
          expr->WriteClipsExpr(stream);
          expr = expr->next;
          if (expr) stream << " ";
        }
        stream << ")\n\n";
      }
    }
  }
}

void ClipsExpr::WriteClipsExpr(ostream& stream)    // Write as any other subexpression
{
  switch (type)
  {
    case ClipsInteger:
    {
      stream << value.integer;
      break;
    }
    case ClipsReal:
    {
      stream << value.real;
      break;
    }
    case ClipsString:
    {
      stream << "\"";
      int i;
      int len = strlen(value.string);
      for (i = 0; i < len; i++)
      {
        char ch = value.string[i];
        if (ch == '"' || ch == '\\')
          stream << "\\";
        stream << ch;
      }

      stream << "\"";
      break;
    }
    case ClipsWord:
    {
      stream << value.word;

      break;
    }
    case ClipsList:
    {
      if (!value.first)
        stream << "()";
      else
      {
        ClipsExpr *expr = value.first;

        stream << "(";
        while (expr)
        {
          expr->WriteClipsExpr(stream);
          expr = expr->next;
          if (expr) stream << " ";
        }
        stream << ")";
      }
      break;
    }
   case ClipsUnknown:
   case ClipsNull: break;
  }
}

void ClipsExpr::WriteClause(void)
{
#ifdef wx_x
  WriteClipsClause(cout);
  cout.flush();
#endif
}

void ClipsExpr::WriteExpr(void)
{
#ifdef wx_x
  WriteClipsExpr(cout);
  cout.flush();
#endif
}

void ClipsExpr::WriteLispExpr(ostream& stream)
{
  switch (type)
  {
    case ClipsInteger:
    {
      stream << value.integer;
      break;
    }
    case ClipsReal:
    {
      stream << value.real;
      break;
    }
    case ClipsString:
    {
      stream << "\"" << value.string << "\"";
      break;
    }
    case ClipsWord:
    {
      stream << value.word;
      break;
    }
    case ClipsList:
    {
      ClipsExpr *expr = value.first;

      stream << "(";
      while (expr)
      {
        expr->WriteLispExpr(stream);
        expr = expr->next;
        if (expr) stream << " ";
      }

      stream << ")";
      break;
    }
   case ClipsUnknown:
   case ClipsNull: break;
  }
}

#ifdef wx_x
void DebugPrintClause(ClipsExpr *expr)
{
  expr->WriteClipsClause(cout);
}
void DebugPrintExpr(ClipsExpr *expr)
{
  expr->WriteClipsExpr(cout);
}
#endif

ClipsExpr MakeList(ClipsExpr *expr ...)
{
  va_list ap;
  ClipsExpr clause(ClipsList);

  va_start(ap, expr);

  if (expr)
  {
    // Flatten list
    if (expr->Type() == ClipsList)
    {
      ClipsExpr *child = expr->GetFirst();
      while (child)
      {
        clause.Append(child);
        child = child->GetNext();
      }
      expr->value.first = NULL;
      expr->last = NULL;
      delete expr;
    }
    else
      clause.Append(expr);
  }

  for (;;)
  {
    ClipsExpr *object = va_arg(ap, ClipsExpr *);
    if (((int)object) == 0)
      break;
    else
    {
      // Flatten list
      if (expr->Type() == ClipsList)
      {
        ClipsExpr *child = expr->GetFirst();
        while (child)
        {
          clause.Append(child);
          child = child->GetNext();
        }
        expr->value.first = NULL;
        expr->last = NULL;
        delete expr;
      }
      else
        clause.Append(object);
    }
  }
  va_end(ap);
  return clause;
}

Bool evenp(float arg)
{
  return (2*((long)(arg/2)) == (long)arg);
}

Bool oddp(float arg)
{
  return (2*((long)(arg/2)) != (long)arg);
}

ClipsExpr StringExplode(wxString& s)
{
  ClipsExpr l;
  char *data = s.GetData();
  int len = strlen(data);
  char explodeBuf[400];
  explodeBuf[0] = 0;
  int i = 0;

  while (i < len)
  {
    int ch = data[i];

    if (isdigit(ch))
    {
      int j = 0;
      Bool hasPoint = FALSE;
      while ((isdigit(data[i]) || data[i] == '.') && (i < len))
      {
        if (data[i] == '.') hasPoint = TRUE;
        
        explodeBuf[j] = data[i];
        i ++;
        j ++;
      }
      explodeBuf[j] = 0;
      if (hasPoint)
      {
        l.Append(new ClipsExpr((float)atof(explodeBuf)));
      }
      else
      {
        l.Append(new ClipsExpr((long)atol(explodeBuf)));
      }
    }
    else if (ch == '"')
    {
      int j = 0;
      i ++;
      while (data[i] != '"' && (i < len))
      {
        explodeBuf[j] = data[i];
        i ++;
        j ++;
      }
      i ++;
      explodeBuf[j] = 0;
      wxString str(explodeBuf);
      l.Append(new ClipsExpr(str));
    }
    else if (ch == ' ')
    {
      while ((i < len) && data[i] == ' ')
        i ++;
    }
    else
    {
      int j = 0;
      while ((data[i] != ' ') && (i < len))
      {
        explodeBuf[j] = data[i];
        i ++;
        j ++;
      }
      explodeBuf[j] = 0;
      wxString str(explodeBuf);
      l.Append(new ClipsExpr(str));
    }
  }
  return l;
}

/*
 * Stream list
 *
 */
 
ClipsStreamList::ClipsStreamList(void):wxList(wxKEY_STRING)
{
}

ClipsStreamList::~ClipsStreamList(void)
{
  CloseAll();
}

Bool ClipsStreamList::OpenStream(char *file, char *name, char *mode)
{
  CloseStream(name);
  if (strcmp(name, "stdout") == 0)
    return TRUE;
    
  FILE *fd = fopen(file, mode);
  if (fd)
  {
    Append(name, (wxObject *)fd);
    return TRUE;
  }
  else return FALSE;
}

Bool ClipsStreamList::CloseStream(char *name)
{
  FILE *fd = FindStream(name);
  if (fd)
  {
    fclose(fd);
    return TRUE;
  }
  else return FALSE;
}

void ClipsStreamList::CloseAll(void)
{
  wxNode *node = First();
  while (node)
  {
    FILE *fd = (FILE *)node->Data();
    fclose(fd);
    delete node;
    node = First();
  }
}

FILE *ClipsStreamList::FindStream(char *name)
{
  wxNode *node = Find(name);
  if (node)
  {
    return (FILE *)node->Data();
  }
  else
    return NULL;
}

wxString ClipsFormat(wxString& sname, char *fmt ...)
{
  char *name = sname.GetData();
  va_list ap;
  static char buffer[512];
  for (int i = 0; i < 512; i++)
    buffer[i] = 0;

  va_start(ap, fmt);

  vsprintf(buffer,fmt,ap) ;

  va_end(ap);

  if (strcmp(name, "nil") == 0)
  {
  }
  else if (strcmp(name, "wclips") == 0 ||
      strcmp(name, "stdout") == 0 ||
      strcmp(name, "t") == 0 ||
      strcmp(name, "wdialog") == 0 ||
      strcmp(name, "wdisplay") == 0 ||
      strcmp(name, "werror") == 0 ||
      strcmp(name, "wwarning") == 0 ||
      strcmp(name, "wtrace") == 0)
  {
    ClipsErrorFunction(buffer);
  }
  else
  {
    FILE *fd = TheClipsStreamList.FindStream(name);
    if (fd)
    {
      fprintf(fd, "%s", buffer);
    }
  }
  return wxString(buffer);
}

// Translate CLIPS %n to newline
char *TranslateFormat(wxString& s)
{
  static char buffer[512];
  buffer[0] = 0;
  char *data = s.GetData();
  int len = strlen(data);
  int i = 0;
  int j = 0;
  while (i < len)
  {
    if (data[i] == '%' && (data[i+1] == 'n'))
    {
      buffer[j] = '\n';
      j ++;
      i += 2;
    }
    else
    {
      buffer[j] = data[i];
      j ++;
      i ++;
    }
  }
  buffer[j] = 0;
  return buffer;
}

/*
 * Stream functions
 */

Bool ClipsOpen(wxString& filename, wxString& name, wxString& mode)
{
  return TheClipsStreamList.OpenStream(filename.GetData(), name.GetData(), mode.GetData());
}

Bool ClipsClose(wxString& name)
{
  if (strlen(name.GetData()) != 0)
    return TheClipsStreamList.CloseStream(name.GetData());
  else
  {
    TheClipsStreamList.CloseAll();
    return TRUE;
  }
}

void ClipsPrintout(wxString& sname, ClipsExpr& expr)
{
  char *name = sname.GetData();
  FILE *fd = NULL;

  if (strcmp(name, "wclips") == 0 ||
      strcmp(name, "stdout") == 0 ||
      strcmp(name, "t") == 0 ||
      strcmp(name, "wdialog") == 0 ||
      strcmp(name, "wdisplay") == 0 ||
      strcmp(name, "werror") == 0 ||
      strcmp(name, "wwarning") == 0 ||
      strcmp(name, "wtrace") == 0)
  {
    fd = NULL;
  }
  else
    fd = TheClipsStreamList.FindStream(name);
  ClipsExpr *element = expr.GetFirst();
  static char buffer[512];
  while (element)
  {
    buffer[0] = 0;
    switch (element->Type())
    {
      case ClipsInteger:
      {
        sprintf(buffer, "%ld", element->IntegerValue());
        break;
      }
      case ClipsReal:
      {
        sprintf(buffer, "%.2f", element->RealValue());
        break;
      }
      case ClipsString:
      case ClipsWord:
      case ClipsUser:
      {
        sprintf(buffer, "%s", element->StringValue());
        break;
      }
      case ClipsStringClass:
      {
        sprintf(buffer, "%s", element->StringClassValue().GetData());
        break;
      }
      default:
        break;
    }
    if (fd)
      fprintf(fd, "%s", buffer);
    else ClipsErrorFunction(buffer);
    element = element->GetNext();
  }
}

wxString ClipsReadLine(wxString& sname)
{
  char *name = sname.GetData();
  static char buffer[512];
  FILE *fd = TheClipsStreamList.FindStream(name);
  if (!fd)
    return wxString("EOF");
  else
  {
    int i = 0;
    int ch = -2;
    while (ch != EOF && ch != '\n')
    {
      ch = getc(fd);
      if (ch != EOF && ch != '\n')
      {
        buffer[i] = ch;
        i ++;
      }
    }
    buffer[i] = 0;
  }
  return wxString(buffer);
}

long ClipsReadInteger(wxString& sname)
{
  char *name = sname.GetData();
  FILE *fd = TheClipsStreamList.FindStream(name);
  if (!fd)
    return 0;
  else
  {
    long ans = 0;
    fscanf(fd, "%ld", &ans);
    return ans;
  }
  return 0;
}

float ClipsReadReal(wxString& sname)
{
  char *name = sname.GetData();
  FILE *fd = TheClipsStreamList.FindStream(name);
  if (!fd)
    return 0.0;
  else
  {
    float ans = 0.0;
    fscanf(fd, "%.2f", &ans);
    return ans;
  }
  return 0.0;
}

wxString ClipsReadString(wxString& sname)
{
  char *name = sname.GetData();
  static char buffer[512];
  FILE *fd = TheClipsStreamList.FindStream(name);
  if (!fd)
    return wxString("EOF");
  else
  {
    fscanf(fd, "%s", buffer);
    return wxString(buffer);
  }
  return wxString("EOF");
}

float ClipsMin(ClipsExpr& expr)
{
  float minValue = 999999.0;
  ClipsExpr *child = expr.GetFirst();
  while (child)
  {
    float val = minValue;
    if (child->Type() == ClipsInteger)
      val = (float)child->IntegerValue();
    else if (child->Type() == ClipsReal)
      val = child->RealValue();
    if (val < minValue)
      minValue = val;
    child = child->GetNext();
  }
  return minValue;
}

float ClipsMax(ClipsExpr& expr)
{
  float maxValue = -999999.0;
  ClipsExpr *child = expr.GetFirst();
  while (child)
  {
    float val = maxValue;
    if (child->Type() == ClipsInteger)
      val = (float)child->IntegerValue();
    else if (child->Type() == ClipsReal)
      val = child->RealValue();
    if (val > maxValue)
      maxValue = val;
    child = child->GetNext();
  }
  return maxValue;
}

static long ClipsGensymId = 0;
wxString ClipsGensym(void)
{
  static char buffer[30];
  sprintf(buffer, "gen%ld", ClipsGensymId);
  ClipsGensymId ++;
  return wxString(buffer);
}

void ClipsSetGen(long g)
{
  ClipsGensymId = g;
}

wxString ClipsGetType(float any)
{
  return wxString("FLOAT");
}

wxString ClipsGetType(wxString& any)
{
  return wxString("STRING");
}

wxString ClipsGetType(long any)
{
  return wxString("INTEGER");
}

wxString ClipsGetType(ClipsExpr& any)
{
  return wxString("MULTIVALUE");  // Doesn't exist I think...
}

long ClipsGetTime(void)
{
#if defined(SVR4) || defined(SYSV)
  struct timeval tp;
  gettimeofday(&tp);
  return (long)tp.tv_sec;
#else
  struct timeb tp;
  ftime(&tp);
  return (long)tp.time;
#endif
}

