/*
 * File:	clipslib.h
 * Purpose:	Run-time support for CLIPS2C-generated C++ code
 * Author:	Julian Smart
 * Created:	1993
 * Updated:	
 * Copyright:	(c) 1993, AIAI, University of Edinburgh
 */

/* sccsid[] = "%W% %G%" */

#ifndef __clipslibh__
#define __clipslibh__

#include <stdio.h>
#include <iostream.h>
#include <wx_list.h>
#include "wx_utils.h"
#include "wxstring.h"

typedef enum {
    ClipsNull,
    ClipsInteger,
    ClipsReal,
    ClipsWord,
    ClipsString,
    ClipsStringClass,
    ClipsFunction,
    ClipsList,
    ClipsUnknown,
    ClipsUser,
    ClipsAny
} ClipsType;

class ClipsExpr
{
 public:
  wxObject *client_data;
  ClipsType returnType;
  ClipsType type;
  union {
    long integer;
    char *word;
    char *string;
    wxString *stringClass;
    float real;
    ClipsExpr *first;  // If is a list expr, points to the first node
    } value;

  ClipsExpr *next;     // If this is a node in a list, points to the next node
  ClipsExpr *last;     // If is a list expr, points to the last node

  ClipsExpr(ClipsExpr& expr);
  ClipsExpr(char *functor);      // Assume this is a new clause - pass functor
  ClipsExpr(ClipsType the_type = ClipsUnknown, char *word_or_string = NULL, Bool allocate = TRUE);
  ClipsExpr(long the_integer);
  ClipsExpr(float the_real);
  ClipsExpr(wxList *the_list);
  ClipsExpr(wxString& string);
  ~ClipsExpr(void);

  // For use of generated C++ code
  void operator = (ClipsExpr& expr);
  float NthReal(int arg);
  long NthInteger(int arg);
  wxString NthString(int arg);
  ClipsExpr NthList(int arg);
  int Member(float arg);
  int Member(long arg);
  int Member(wxString& arg);
  ClipsExpr DeleteRange(int from, int to);
  wxString Implode(void);
  ClipsExpr Subseq(int from, int to);
  ClipsExpr Replace(int from, int to, ClipsExpr& expr);
  ClipsExpr Replace(int from, int to, float arg);
  ClipsExpr Replace(int from, int to, long arg);
  ClipsExpr Replace(int from, int to, wxString& arg);
  ClipsExpr Insert(int pos, ClipsExpr& expr);
  ClipsExpr Insert(int pos, float arg);
  ClipsExpr Insert(int pos, long arg);
  ClipsExpr Insert(int pos, wxString& arg);
  ClipsExpr FirstList(void);
  ClipsExpr Rest(void);

  int Number(void);

  inline ClipsType Type(void) { return type; }

  inline ClipsType GetReturnType(void) { return returnType; }
  inline void SetReturnType(ClipsType typ) { returnType = typ; }

  inline long IntegerValue(void)
  {
    if (type == ClipsInteger)
      return value.integer;
    else if (type == ClipsReal)
      return (long)value.real;
    else return 0;
  }

  inline float RealValue(void) {
    if (type == ClipsReal)
      return value.real;
    else if (type == ClipsInteger)
      return (float)value.integer;
    else return 0.0;
  }

  inline char *WordValue(void) {
    if (type == ClipsWord)
      return value.word;
    else if (type == ClipsString)
      return value.string;
    else return NULL;
  }

  inline char *StringValue(void) {
    if (type == ClipsString)
      return value.string;
    else if (type == ClipsWord)
      return value.word;
    else return NULL;
  }

  inline char *UserValue(void) {
    if (type == ClipsUser)
      return value.string;
    else if (type == ClipsString)
      return value.word;
    else if (type == ClipsWord)
      return value.word;
    else return NULL;
  }

  inline wxString& StringClassValue(void) {
    return *(value.stringClass);
  }

  // Get nth arg of clause (starting from 1)
  ClipsExpr *Arg(ClipsType type, int arg);

  // Return nth argument of a list expression (starting from zero)
  ClipsExpr *Nth(int arg);

  ClipsExpr *GetAttributeValueNode(char *word);  // Use only for a clause or list
  ClipsExpr *AttributeValue(char *word);  // Use only for a clause
  char *Functor(void);                     // Only for a clause
  Bool IsFunctor(char *f);                     // Only for a clause
  void WriteClipsClause(ostream& stream);  // Write this expression as a top-level clause
  void WriteClipsExpr(ostream& stream);    // Write as any other subexpression
  void WriteLispExpr(ostream& stream);

  // Debugging purposes - write to std output
  void WriteClause(void);
  void WriteExpr(void);

  // Append an expression to a list
  void Append(ClipsExpr *expr);
  // Insert at beginning of list
  void Insert(ClipsExpr *expr);

  // Get first expr in list
  inline ClipsExpr *GetFirst(void) { return ((type == ClipsList) ? value.first : NULL); }

  // Get next expr if this is a node in a list
  inline ClipsExpr *GetNext(void) { return next; }

  // Get last expr in list
  inline ClipsExpr *GetLast(void) { return ((type == ClipsList) ? last : NULL); }

  // This should really be called SetAttributeValue since any existing
  // attribute-value is deleted first.
  void AddAttributeValue(char *attribute, long value);
  void AddAttributeValue(char *attribute, float value);
  void AddAttributeValueWord(char *attribute, char *value);
  void AddAttributeValueString(char *attribute, char *value);
  void AddAttributeValue(char *attribute, wxList *value);
  void AddAttributeValue(char *attribute, ClipsExpr *value);
  void AddAttributeValueStringList(char *attribute, wxList *string_list);

  void DeleteAttributeValue(char *attribute);

  // Retrieving values - only assigns values if they were found, and of right type
  // Usage: e.g.
  // int id = 0;
  // clause->AssignAttributeValue("id", &id);

  void AssignAttributeValue(char *att, int *var);
  void AssignAttributeValue(char *att, long *var);
  void AssignAttributeValue(char *att, float *var);
  void AssignAttributeValue(char *att, char **var);  // Word OR string -> string
  void AssignAttributeValue(char *att, ClipsExpr **var);

  // Add string items to list if the list attribute exists
  void AssignAttributeValueStringList(char *att, wxList *var);

  // Associate other data with this expression, e.g. when reading in a
  // number of linked items - store C++ object pointer with the expression
  // so we can index into the Clips database and fish out the pointer.
  inline void SetClientData(wxObject *data) { client_data = data; }
  inline wxObject *GetClientData(void) { return client_data; }
};

/*
 * Run-time support
 *
 */

ClipsExpr MakeList(ClipsExpr *expr ...);

Bool evenp(float arg);
Bool oddp(float arg);

ClipsExpr StringExplode(wxString& s);
wxString ClipsFormat(wxString& name, char *fmt ...);
char *TranslateFormat(wxString& s);
void ClipsPrintout(wxString& name, ClipsExpr& expr);
Bool ClipsOpen(wxString& filename, wxString& name, wxString& mode);
Bool ClipsClose(wxString& name);

wxString ClipsReadLine(wxString& sname);
long ClipsReadInteger(wxString& sname);
float ClipsReadReal(wxString& name);
wxString ClipsReadString(wxString& name);

wxString ClipsGensym(void);
void ClipsSetGen(long g);

#define ClipsMakeReal(x) ((float)x)
#define ClipsMakeInteger(x) ((long)x)
#define ClipsRound(x) ((long)(x + 0.5))

float ClipsMin(ClipsExpr& expr);
float ClipsMax(ClipsExpr& expr);

wxString ClipsGetType(float any);
wxString ClipsGetType(wxString& any);
wxString ClipsGetType(long any);
wxString ClipsGetType(ClipsExpr& any);

long ClipsGetTime(void);

/*
 * I/O
 *
 */

class ClipsStreamList: public wxList
{
 public:
  ClipsStreamList(void);
  ~ClipsStreamList(void);

  Bool OpenStream(char *file, char *name, char *mode);
  Bool CloseStream(char *name);
  void CloseAll(void);
  FILE *FindStream(char *name);
};

#endif // __clipslibh__


