/*
 * File:	gcool.cc
 * Purpose:	Interpretation and generation of COOL constructs
 * Author:	Julian Smart
 * Created:	1993
 * Updated:	
 * Copyright:	(c) 1993, AIAI, University of Edinburgh
 */

static const char sccsid[] = "%W% %G%";

#include <stdarg.h>
#include <ctype.h>
#include <iostream.h>
#include <fstream.h>

#include "wx_hash.h"
#include "wxstring.h"
#include "clips2c.h"

extern Bool wxYield(void);
extern char *FileNameFromPath(char *);
extern char *PathOnly(char *);
extern Bool FileExists(char *);

extern void Indent(ostream& stream, int indent);

Bool ClipsDatabase::ProcessMessageHandler(ClipsExpr *expr)
{
  ClipsExpr *funcNameExpr = expr->Nth(1);
  char *funcName = funcNameExpr->WordValue();
  ClipsExpr *classNameExpr = expr->Nth(2);
  char *className = classNameExpr->WordValue();
  ClipsExpr *funcArgs = expr->Nth(3);
  ClipsConstruct *construct =
        TheNameSpace.AddConstruct(funcName, CONSTRUCT_TYPE_HANDLER, expr, className);
  ConstructList.Append(construct);

  ClipsConstruct *classConstruct = TheNameSpace.FindConstruct(className);
  if (!classConstruct)
  {
    char buf[200];
    sprintf(buf, "Cannot find class for handler %s:%s", funcName, className);
    ClipsError(buf);
  }
  else classConstruct->members.Append(construct);
  
  // Construct a local variable list, initialized with the
  // arguments
  ClipsExpr *arg = funcArgs->GetFirst();
  while (arg)
  {
    char *val = arg->WordValue();
    char *argName = val;
    if (val[0] == '?') argName = (char *)(val + 1);
    
    construct->locals.AddConstruct(argName, CONSTRUCT_TYPE_PARAMETER);
    arg = arg->GetNext();
  }

  construct->returnType = ClipsNull;
  ClipsExpr *body = expr->Nth(3);
  while (body)
  {
    ClipsType tmp;
    ProcessExpression(construct, body, &tmp, 0);
    body = body->GetNext();
  }
  return TRUE;
}

Bool ClipsDatabase::ProcessClass(ClipsExpr *expr)
{
  ClipsExpr *className = expr->Nth(1);
  ClipsExpr *component = expr->Nth(2);
  ClipsConstruct *construct = new ClipsConstruct(className->WordValue(), CONSTRUCT_TYPE_CLASS, expr);
  if (TheNameSpace.Get(className->WordValue()))
  {
    TheNameSpace.Delete(className->WordValue());
  }
  TheNameSpace.Put(className->WordValue(), construct);
  ConstructList.Append(construct);

  char *parentClass = NULL;

  while (component)
  {
    if (component->Type() == ClipsList)
    {
      ClipsExpr *keywordExpr = component->Nth(0);
      if (keywordExpr->Type() == ClipsWord)
      {
        char *keyword = keywordExpr->WordValue();
        if (strcmp(keyword, "is-a") == 0)
        {
          ClipsExpr *classParentExpr = component->Nth(1);
          if (classParentExpr->Type() == ClipsWord)
          {
            parentClass = classParentExpr->WordValue();
            construct->SetParentClass(parentClass);
          }
        }
        else if (strcmp(keyword, "slot") == 0)
        {
          ClipsExpr *slotExpr = component->Nth(1);
          ClipsConstruct *slot = new ClipsConstruct(NULL, CONSTRUCT_TYPE_SLOT);
          construct->members.Append(slot);
          
          if (slotExpr->Type() == ClipsWord)
          {
            slot->SetName(slotExpr->WordValue());
          }
          
          // Process facets
          ClipsExpr *facetExpr = component->Nth(2);
          while (facetExpr)
          {
            if (facetExpr->Type() == ClipsList)
            {
              ClipsExpr *facetNameExpr = facetExpr->Nth(0);
              if (facetNameExpr->Type() == ClipsWord)
              {
                char *facetName = facetNameExpr->WordValue();
                if (strcmp(facetName, "type") == 0)
                {
                  ClipsExpr *typeExpr = facetExpr->Nth(1);
                  if (typeExpr->Type() == ClipsWord)
                  {
                    char *typeName = typeExpr->WordValue();
                    if (strcmp(typeName, "STRING") == 0)
                      slot->SetReturnType(ClipsStringClass);
                    else if (strcmp(typeName, "NUMBER") == 0)
                      slot->SetReturnType(ClipsReal);
                    else if (strcmp(typeName, "SYMBOL") == 0)
                      slot->SetReturnType(ClipsStringClass);
                    else if (strcmp(typeName, "LIST") == 0) // ???
                      slot->SetReturnType(ClipsList);
                    else
                    {
                      slot->SetReturnType(ClipsUser);
                      slot->SetUserType(typeName);
                    }
		  }
                }
                else if (strcmp(facetName, "default") == 0 || strcmp(facetName, "default-dynamic") == 0)
                {
                  ClipsExpr *defaultExpr = facetExpr->Nth(1);
                  slot->SetValue(defaultExpr);
                }
              }
            }
            facetExpr = facetExpr->GetNext();
          }
        }
      }
    }
    component = component->GetNext();
  }
  AddAccessorFunctions(construct);
  return TRUE;
}

Bool ClipsDatabase::AddAccessorFunctions(ClipsConstruct *construct)
{
  char *className = construct->GetName();
  wxNode *node = construct->members.First();
  while (node)
  {
    ClipsConstruct *slot = (ClipsConstruct *)node->Data();
    if (slot->GetType() == CONSTRUCT_TYPE_SLOT)
    {
    char buf[200];
    char *slotName = slot->GetName();

    // get- accessor
    strcpy(buf, "get-");
    strcat(buf, slotName);
    ClipsConstruct *func = TheNameSpace.AddConstruct(buf, CONSTRUCT_TYPE_HANDLER, NULL, className);
    func->SetReturnType(slot->GetReturnType());
    ConstructList.Append(func);
    // Construct the 'code'
    ClipsExpr *funcBody = new ClipsExpr(ClipsList);
    funcBody->Append(new ClipsExpr(ClipsWord, "defmessage-handler"));
    funcBody->Append(new ClipsExpr(ClipsWord, className));
    funcBody->Append(new ClipsExpr(ClipsWord, buf));
    funcBody->Append(new ClipsExpr(ClipsList));
    ClipsExpr *retExpr = new ClipsExpr(ClipsList);
    retExpr->Append(new ClipsExpr(ClipsWord, "return"));
    char varBuf[150];
    strcpy(varBuf, "?");
    strcat(varBuf, slotName);
    retExpr->Append(new ClipsExpr(ClipsWord, varBuf));
    funcBody->Append(retExpr);
    func->SetValue(funcBody);
    Insert((wxObject *)funcBody); // Append to database so it gets deleted eventually.
    construct->members.Append(func);

    // put- accessor
    strcpy(buf, "put-");
    strcat(buf, slotName);
    func = TheNameSpace.AddConstruct(buf, CONSTRUCT_TYPE_HANDLER, NULL, className);
    func->SetReturnType(ClipsNull);
    ClipsConstruct *local = func->locals.AddConstruct("arg", CONSTRUCT_TYPE_PARAMETER);
    local->SetReturnType(slot->GetReturnType());
    ConstructList.Append(func);
    // Construct the 'code'
    funcBody = new ClipsExpr(ClipsList);
    funcBody->Append(new ClipsExpr(ClipsWord, "defmessage-handler"));
    funcBody->Append(new ClipsExpr(ClipsWord, className));
    funcBody->Append(new ClipsExpr(ClipsWord, buf));
    ClipsExpr *argList = new ClipsExpr(ClipsList);
    argList->Append(new ClipsExpr(ClipsWord, "?arg"));
    funcBody->Append(argList);
    ClipsExpr *bindExpr = new ClipsExpr(ClipsList);
    bindExpr->Append(new ClipsExpr(ClipsWord, "bind"));
    strcpy(varBuf, "?");
    strcat(varBuf, slotName);
    bindExpr->Append(new ClipsExpr(ClipsWord, varBuf));
    bindExpr->Append(new ClipsExpr(ClipsWord, "?arg"));
    funcBody->Append(bindExpr);
    func->SetValue(funcBody);
    Insert((wxObject *)funcBody); // Append to database so it gets deleted eventually.
    construct->members.Append(func);
    }
    
    node = node->Next();
  }
  return TRUE;
}

Bool ClipsDatabase::GenerateMessageHandler(ClipsConstruct *construct, ostream& stream)
{
  return TRUE;
}

Bool ClipsDatabase::GenerateClass(ClipsConstruct *construct, ostream& stream)
{
  char className[150];
  strcpy(className, construct->GetCName());
  stream << "class " << className;
  char *parentClass = NULL;
  if (parentClass = construct->GetParentClass())
  {
    if (strcmp(parentClass, "USER") == 0)
      stream << ": public wxObject";
    else
    {
      strcpy(className, parentClass);
      ClipsNameToC(className);
      stream << ": public " << className;
    }
  }
  stream << "\n{\n";
  stream << " public:\n";

  // Generate slots (member variables)
  wxNode *node = construct->members.First();
  while (node)
  {
    ClipsConstruct *slot = (ClipsConstruct *)node->Data();
    if (slot->GetType() == CONSTRUCT_TYPE_SLOT)
    {
    static char buf[100];
    buf[0] = 0;
    switch (slot->GetReturnType())
    {
      case ClipsReal:
      {
        strcpy(buf, "float");
        break;
      }
      case ClipsInteger:
      {
        strcpy(buf, "long");
        break;
      }
      case ClipsString:
      case ClipsStringClass:
      case ClipsWord:
      {
        strcpy(buf, "wxString");
        break;
      }
      case ClipsList:
      {
        strcpy(buf, "ClipsExpr");
        break;
      }
      case ClipsUser:
      {
        strcpy(buf, "wxString");
        break;
      }
      default:
      {
        strcpy(buf, "unknown");
        char buf2[200];
        sprintf(buf2, "Unknown slot type in class definition %s", construct->GetName());
        ClipsError(buf2);
        break;
      }
    }
    stream << "  " << buf << " ";
    char *slotName = slot->GetCName();
    stream << slotName << ";\n";
    }
    node = node->Next();
  }
  stream << "\n";

  // Generate constructor
  stream << "  " << className << "(void)\n  {\n";
  node = construct->members.First();
  while (node)
  {
    ClipsConstruct *slot = (ClipsConstruct *)node->Data();
    if (slot->GetType() == CONSTRUCT_TYPE_SLOT)
    {
    char slotName[100];
    strcpy(slotName, slot->GetCName());
    ClipsType tmp;

    ClipsExpr *defaultValue = slot->GetValue();

    switch (slot->GetReturnType())
    {
      case ClipsReal:
      {
        stream << "    " << slotName << " = ";
        if (defaultValue)
          GenerateExpression(construct, defaultValue, &tmp, 2, stream);
        else
          stream << "0.0";
        stream << ";\n";
        break;
      }
      case ClipsInteger:
      {
        stream << "    " << slotName << " = ";
        if (defaultValue)
          GenerateExpression(construct, defaultValue, &tmp, 2, stream);
        else
          stream << "0";
        stream << ";\n";
        break;
      }
      case ClipsString:
      case ClipsStringClass:
      case ClipsWord:
      case ClipsList:
      case ClipsUser:
      {
        if (defaultValue)
        {
          stream << "    " << slotName << " = ";
          GenerateExpression(construct, defaultValue, &tmp, 2, stream);
          stream << ";\n";
        }
        break;
      }
      default:
      {
        break;
      }
    }
    }
    node = node->Next();
  }
  stream << "  }\n";
  

  stream << "};\n\n";
  return TRUE;
}

