/*
 * File:	gfunc.cc
 * Purpose:	Interpretation and generation of function-related CLIPS constructs,
 *              and global variables. See gcool.cc for COOL-related generation.
 * Author:	Julian Smart
 * Created:	1993
 * Updated:	
 * Copyright:	(c) 1993, AIAI, University of Edinburgh
 */

static const char sccsid[] = "%W% %G%";

#include <stdarg.h>
#include <ctype.h>
#include <iostream.h>
#include <fstream.h>

#include "wx_hash.h"
#include "wx_utils.h"
#include "wxstring.h"
#include "clips2c.h"

extern Bool wxYield(void);
/*
extern char *FileNameFromPath(char *);
extern char *PathOnly(char *);
extern Bool FileExists(const char *);
*/

extern Bool generateMode;

void Indent(ostream& stream, int indent);

/*
 * Code generation
 *
 */

Bool ClipsDatabase::GenerateC(char *filename)
{
  ConstructList.Clear();
  TheNameSpace.ClearConstructs();

  if (!Phase1())
    return FALSE;

  ofstream stream(filename);
  return Phase2(stream, filename);
}
 
// Phase 1: traverse CLIPS code setting up tables of variables and functions, etc.
Bool ClipsDatabase::Phase1(void)
{
  wxNode *node = First();
  while (node)
  {
    wxYield();
    ClipsExpr *expr = (ClipsExpr *)node->Data();
    if (expr->Type() == ClipsList)
    {
      if ((expr->Nth(0) != NULL) && (expr->Nth(0)->Type() == ClipsWord))
      {
        char *theWord = expr->Nth(0)->WordValue();
        if (strcmp(theWord, "deffunction") == 0)
        {
          ProcessFunction(expr);
        }
        else if (strcmp(theWord, "defmessage-handler") == 0)
        {
          ProcessMessageHandler(expr);
        }
        else if (strcmp(theWord, "defclass") == 0)
        {
          ProcessClass(expr);
        }
        else if (strcmp(theWord, "defglobal") == 0)
        {
          ProcessGlobal(expr);
        }
        else if (strcmp(theWord, "load") == 0)
        {
          // Insert in front of the next construct.
          ProcessLoad(expr, node->Next());
        }
        else if (strncmp(theWord, "def", 3) == 0)
        {
          char buf[100];
          sprintf(buf, "%s is an unrecognised construct, sorry.", theWord);
          ClipsError(buf);
        }
        else
        {
          ProcessStatement(expr);
        }
      }
    }
    node = node->Next();
  }
  return TRUE;
}

// Phase 2: resolve variable types and output code.
Bool ClipsDatabase::Phase2(ostream& stream, char *toFilename)
{
  // Generate header
  stream << "/*\n";
  stream << " * " << FileNameFromPath(toFilename) << "\n";
  stream << " * File generated by Clips2C from " << FileNameFromPath(GetInputFile()) << "\n";
  if (GetGenerationMode() == MODE_WXCLIPS)
    stream << " * for wxCLIPS compatibility\n";
  if (GetGenerationMode() == MODE_HARDY)
    stream << " * for HARDY compatibility\n";
  stream << " */\n\n";
  
//  stream << "#include <wx.h>\n";
  stream << "#include \"math.h\"\n";
  stream << "#include \"stdlib.h\"\n";
  stream << "#include \"wxstring.h\"\n";
  stream << "#include \"clipslib.h\"\n";
  if (GetGenerationMode() == MODE_WXCLIPS || GetGenerationMode() == MODE_HARDY)
    stream << "#include \"wx_cmds.h\"\n";
  if (GetGenerationMode() == MODE_HARDY)
    stream << "#include \"hapi.h\"\n";
  stream << "\n";
  
  GeneratePrototypes(stream);

  stream << "\n/*\n * Global variables\n */\n\n";
  wxNode *node = ConstructList.First();
  while (node)
  {
    wxYield();
    ClipsConstruct *construct = (ClipsConstruct *)node->Data();
    switch (construct->GetType())
    {
      case CONSTRUCT_TYPE_GLOBAL:
      {
        GenerateGlobal(construct, stream);
        break;
      }
      default:
        break;
    }
    node = node->Next();
  }

  stream << "\n/*\n * Classes\n */\n\n";

  node = ConstructList.First();
  while (node)
  {
    wxYield();
    ClipsConstruct *construct = (ClipsConstruct *)node->Data();
    switch (construct->GetType())
    {
      case CONSTRUCT_TYPE_CLASS:
      {
        GenerateClass(construct, stream);
        break;
      }
      default:
        break;
    }
    node = node->Next();
  }

  stream << "\n/*\n * Functions\n */\n\n";

  node = ConstructList.First();
  while (node)
  {
    wxYield();
    ClipsConstruct *construct = (ClipsConstruct *)node->Data();
    switch (construct->GetType())
    {
      case CONSTRUCT_TYPE_FUNCTION:
      {
        GenerateFunction(construct, stream);
        break;
      }
      case CONSTRUCT_TYPE_HANDLER:
      {
        GenerateMessageHandler(construct, stream);
        break;
      }
      default:
        break;
    }
    node = node->Next();
  }

  // Now generate top-level statements, if any.
  stream << "/*\n * Top-level statements\n */\n\n";

  node = ConstructList.First();
  stream << "void ExecuteClipsStatements(void)\n";
  stream << "{\n";
  while (node)
  {
    wxYield();
    ClipsConstruct *construct = (ClipsConstruct *)node->Data();
    switch (construct->GetType())
    {
      case CONSTRUCT_TYPE_STATEMENT:
      {
        GenerateStatement(construct, stream);
        break;
      }
      default:
        break;
    }
    node = node->Next();
  }
  stream << "}\n\n";
  if (GetGenerationMode() == MODE_HARDY)
  {
    stream << "/*\n * Initialization function called by HARDY\n */\n\n";
    stream << "void gyInitializeCustomCode(void)\n";
    stream << "{\n";
    stream << "  ExecuteClipsStatements();\n";
    stream << "}\n";
  }
  return TRUE;
}

Bool ClipsDatabase::ProcessFunction(ClipsExpr *expr)
{
  ClipsExpr *funcName = expr->Nth(1);
  ClipsExpr *funcArgs = expr->Nth(2);
  ClipsConstruct *construct = new ClipsConstruct(funcName->WordValue(), CONSTRUCT_TYPE_FUNCTION, expr);
  if (TheNameSpace.Get(funcName->WordValue()))
  {
    TheNameSpace.Delete(funcName->WordValue());
  }
  TheNameSpace.Put(funcName->WordValue(), construct);
  ConstructList.Append(construct);

  // Construct a local variable list, initialized with the
  // arguments
  ClipsExpr *arg = funcArgs->GetFirst();
  while (arg)
  {
    char *val = arg->WordValue();
    char *argName = val;
    if (val[0] == '?') argName = (char *)(val + 1);
    
    construct->locals.AddConstruct(argName, CONSTRUCT_TYPE_PARAMETER);
    arg = arg->GetNext();
  }

  construct->returnType = ClipsNull;
  ClipsExpr *body = expr->Nth(3);
  while (body)
  {
    ClipsType tmp;
    ProcessExpression(construct, body, &tmp, 0);
    body = body->GetNext();
  }
  
  return TRUE;
}

Bool ClipsDatabase::ProcessGlobal(ClipsExpr *expr)
{
  ClipsExpr *globName = expr->Nth(1);
  while (globName)
  {
    char *name = globName->WordValue();
    if (name[0] == '?')
      name = name + 1;

    // Ignore equals
    globName = globName->GetNext();

    // Get value
    if (globName)
      globName = globName->GetNext();
      
    if (globName)
    {
      ClipsConstruct *construct = new ClipsConstruct(name, CONSTRUCT_TYPE_GLOBAL, expr);

      ClipsType returnType = ClipsUnknown;

      ProcessExpression(construct, globName, &returnType, 0);

      construct->returnType = returnType;

      TheNameSpace.Put(name, construct);
      ConstructList.Append(construct);
      globName = globName->GetNext();
    }
  }
  
  return TRUE;
}

Bool ClipsDatabase::ProcessStatement(ClipsExpr *expr)
{
  ClipsConstruct *construct = new ClipsConstruct("<statement>", CONSTRUCT_TYPE_STATEMENT, expr);

  ClipsType returnType = ClipsUnknown;
  ProcessExpression(construct, expr, &returnType, 0);

  construct->returnType = ClipsNull;

  ConstructList.Append(construct);
  return TRUE;
}

Bool ClipsDatabase::ProcessLoad(ClipsExpr *expr, wxNode *pos)
{
  ClipsExpr *fileExpr = expr->Nth(1);
  if (!fileExpr || (fileExpr->Type() != ClipsString)) return FALSE;
  char *fileName = fileExpr->StringValue();
  char buf[300];
  strcpy(buf, fileName);
  if (!FileExists(buf))
  {
    strcpy(buf, PathOnly(GetInputFile()));
#ifdef wx_msw
    strcat(buf, "\\");
#else
    strcat(buf, "/");
#endif
    strcat(buf, fileName);
  }
  if (!FileExists(buf))
  {
    char buf2[300];
    sprintf(buf2, "Cannot find file %s to load.", fileName);
    ClipsError(buf2);
    return FALSE;
  }
  return ReadClips(buf, pos);
}
    
// Recursively process the expression in the given function,
// returning an expression. returnType is the actual type of the
// expression, which depends on how it's evaluated (may be a function
// or a variable)
ClipsExpr *ClipsDatabase::ProcessExpression(ClipsConstruct *func, ClipsExpr *expr,
                       ClipsType *returnType, int indent)
{
  *returnType = expr->Type();
  switch (expr->Type())
  {
    case ClipsList:
    {
      if (expr->Nth(0) == NULL)
        return expr;
      ClipsExpr *first = expr->Nth(0);
      if (first->Type() != ClipsWord)
        return expr;
      // Ok, now we have a function, either built-in or user-defined.
      char *name = first->WordValue();
      if (strcmp(name, "if") == 0)
        return ProcessIf(func, expr, returnType, indent);
      else if (strcmp(name, "declare") == 0)
        return ProcessDeclare(func, expr, returnType, indent);
      else if (strcmp(name, "return") == 0)
        return ProcessReturn(func, expr, returnType, indent);
      else if (strcmp(name, "while") == 0)
        return ProcessWhile(func, expr, returnType, indent);
      else if (strcmp(name, "progn") == 0 || strcmp(name, "progn$") == 0)
        return ProcessProgn(func, expr, returnType, indent);
      else if (strcmp(name, "switch") == 0)
        return ProcessSwitch(func, expr, returnType, indent);
      else if (strcmp(name, "loop-for-count") == 0)
        return ProcessLoop(func, expr, returnType, indent);
      else if (strcmp(name, "break") == 0)
      {
        *returnType = ClipsNull;
        expr->SetReturnType(ClipsNull);
        return expr;
      }
      else if (strcmp(name, "bind") == 0)
        return ProcessBind(func, expr, returnType, indent);
      else if (strcmp(name, "eq") == 0)
        return ProcessEq(func, expr, returnType, indent);
      else if (strcmp(name, "neq") == 0)
        return ProcessNeq(func, expr, returnType, indent);
      else if (strcmp(name, "str-cat") == 0 || strcmp(name, "sym-cat") == 0)
        return ProcessStrcat(func, expr, returnType, indent);
      else if (strcmp(name, "mv-append") == 0 || strcmp(name, "create$") == 0 || strcmp(name, "create") == 0)
        return ProcessMvAppend(func, expr, returnType, indent);
      else if (strcmp(name, "member") == 0 || strcmp(name, "member$") == 0)
        return ProcessMember(func, expr, returnType, indent);
      else if (strcmp(name, "length") == 0 || strcmp(name, "length$") == 0)
        return ProcessLength(func, expr, returnType, indent);
      else if (strcmp(name, "format") == 0)
        return ProcessFormat(func, expr, returnType, indent);
      else if (strcmp(name, "printout") == 0)
        return ProcessPrintout(func, expr, returnType, indent);
      else if (strcmp(name, "nth") == 0 || strcmp(name, "nth$") == 0)
        return ProcessNth(func, expr, returnType, indent);
      else if (strcmp(name, "+") == 0 || strcmp(name, "-") == 0 ||
               strcmp(name, "*") == 0 || strcmp(name, "/") == 0)
        return ProcessArithmetic(func, expr, returnType, indent);
      else if (strcmp(name, "min") == 0 || strcmp(name, "max") == 0)
        return ProcessArithmetic(func, expr, returnType, indent);
      else if (strcmp(name, "<") == 0 || strcmp(name, ">") == 0 ||
               strcmp(name, "<=") == 0 || strcmp(name, ">=") == 0 ||
               strcmp(name, "=") == 0 ||  strcmp(name, "<>") == 0)
        return ProcessComparison(func, expr, returnType, indent);
      else if (strcmp(name, "numberp") == 0 || strcmp(name, "floatp") == 0 ||
               strcmp(name, "integerp") == 0 || strcmp(name, "lexemep") == 0 ||
               strcmp(name, "stringp") == 0 || strcmp(name, "symbolp") == 0 ||
               strcmp(name, "evenp") == 0 || strcmp(name, "oddp") == 0 ||
               strcmp(name, "multifieldp") == 0)
        return ProcessPredicate(func, expr, returnType, indent);
      else
      {
        ClipsConstruct *clipsFunc = TheNameSpace.FindConstruct(name);
        if (clipsFunc && (clipsFunc->GetType() == CONSTRUCT_TYPE_FUNCTION))
          return ProcessClipsFunction(func, expr, returnType, indent);
        else
          return ProcessUserFunction(func, expr, returnType, indent);
      }
      break;
    }
    case ClipsWord:
    {
      // Check for variable
      char *name = expr->WordValue();
      if (name[0] == '?')
      {
        name = name + 1;
        ClipsConstruct *construct = func->locals.FindConstruct(name);
        if (!construct)
          construct = TheNameSpace.FindConstruct(name);
        if (construct)
        {
          *returnType = construct->GetReturnType();
          expr->SetReturnType(*returnType);
        }
        else
          *returnType = ClipsUnknown;
      }
      else
      {
        if (strcmp(name, "TRUE") == 0 || strcmp(name, "FALSE") == 0)
        {
          *returnType = ClipsInteger;
          expr->SetReturnType(*returnType);
        }
      }
      return expr;
      break;
    }
    case ClipsInteger:
    case ClipsReal:
    {
      expr->SetReturnType(expr->Type());
      return expr;
      break;
    }
    case ClipsString:
    {
      expr->SetReturnType(ClipsStringClass);
      return expr;
      break;
    }
    default:
    {
      return expr;
      break;
    }
  }
  return NULL;
}

ClipsExpr *ClipsDatabase::ProcessDeclare(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent)
{
  *returnType = ClipsNull;
  expr->SetReturnType(ClipsNull);

  ClipsExpr *currentExpr = expr->Nth(1);

  while (currentExpr)
  {
    if (!currentExpr || (currentExpr->Type() != ClipsWord))
      return expr;
    char *name = currentExpr->WordValue();
    if (name[0] == '?') name = name + 1;
    ClipsConstruct *construct = func->locals.FindConstruct(name);
    if (!construct)
      construct = TheNameSpace.FindConstruct(name);
    if (!construct)
    {
      char buf[200];
      sprintf(buf, "In function '%s', 'declare' could not find variable %s",
          func->GetName(), name);
      ClipsError(buf);
      return expr;
    }
    currentExpr = currentExpr->GetNext();

    if (!currentExpr)
      return expr;
      
    char *typeS = NULL;
    if (currentExpr->Type() == ClipsWord)
      typeS = currentExpr->WordValue();
    else
      typeS = currentExpr->StringValue();
    ClipsType theType = ClipsUnknown;
    char buf2[30];
    strcpy(buf2, typeS);
    for (int i = 0; i < (int)strlen(typeS); i++)
      buf2[i] = tolower(buf2[i]);
    
    if (strcmp(buf2, "string") == 0 || strcmp(buf2, "word") == 0)
      theType = ClipsString;
    else if (strcmp(buf2, "integer") == 0 || strcmp(buf2, "int") == 0)
      theType = ClipsInteger;
    else if (strcmp(buf2, "real") == 0 || strcmp(buf2, "float") == 0)
      theType = ClipsReal;
    else if (strcmp(buf2, "list") == 0 || strcmp(buf2, "multivalue") == 0 ||
      strcmp(buf2, "multi-value") == 0)
      theType = ClipsList;
    else
    {
      char buf[200];
      sprintf(buf, "In function %s, 'declare' type for %s should be one of integer, string, real or list, not %s",
          func->GetName(), name, typeS);
      ClipsError(buf);
      return expr;
    }
    construct->returnType = theType;
    currentExpr = currentExpr->GetNext();
  }
  return expr;
}

ClipsExpr *ClipsDatabase::ProcessReturn(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent)
{
  *returnType = ClipsNull;
  expr->SetReturnType(ClipsNull);
  ClipsExpr *retExpr = expr->Nth(1);
  if (!retExpr)
    return expr;
  ProcessExpression(func, retExpr, returnType, indent);
  func->returnType = *returnType;
  return expr;
}

ClipsExpr *ClipsDatabase::ProcessIf(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent)
{
  *returnType = ClipsUnknown;
  
  ClipsExpr *thing = expr->GetFirst();
  // Ignore 'if'
  thing = thing->GetNext();

  // Next one is the test
  ClipsType typ;
  ProcessExpression(func, thing, &typ, indent);

  // Next one is 'then'
  thing = thing->GetNext();
  if ((thing->Type() != ClipsWord) || (strcmp(thing->WordValue(), "then") != 0))
  {
    ClipsError("Expected a 'then'");
  }

  // Next one is the 'then' expr
  thing = thing->GetNext();
  ClipsExpr *thenExpr = NULL;
  while (thing && !((thing->Type() == ClipsWord && strcmp(thing->WordValue(), "else") == 0)))
  {
    thenExpr = ProcessExpression(func, thing, returnType, indent);
    thing = thing->GetNext();
  }
  expr->SetReturnType(thenExpr->GetReturnType());

  // Is there an 'else'?  
  if (thing)
  {
    if ((thing->Type() != ClipsWord) || (strcmp(thing->WordValue(), "else") != 0))
    {
      ClipsError("Expected an 'else'");
      return thenExpr;
    }

    // Yes, so...
    thing = thing->GetNext();
    while (thing)
    {
      ClipsExpr *elseExpr = ProcessExpression(func, thing, returnType, indent);
      expr->SetReturnType(elseExpr->GetReturnType());
      thing = thing->GetNext();
    }
    return expr;
  }
  else
    return thenExpr;
}

ClipsExpr *ClipsDatabase::ProcessWhile(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent)
{
  *returnType = ClipsNull;
  ClipsExpr *boolExpr = expr->Nth(1);
  ClipsType typ;
  ProcessExpression(func, boolExpr, &typ, indent);

  ClipsExpr *statement = expr->Nth(2);
  if ((statement->Type() == ClipsWord) && (strcmp(statement->WordValue(), "do") == 0))
    statement = statement->GetNext();

  while (statement)
  {
    ProcessExpression(func, statement, &typ, indent);
    statement = statement->GetNext();
  }
  
  expr->SetReturnType(ClipsNull);
  return expr;
}

ClipsExpr *ClipsDatabase::ProcessProgn(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent)
{
  *returnType = ClipsNull;
  ClipsType typ;
  ClipsExpr *statement = expr->Nth(1);
  while (statement)
  {
    ProcessExpression(func, statement, &typ, indent);
    statement = statement->GetNext();
  }
  
  expr->SetReturnType(ClipsNull);
  return expr;
}

ClipsExpr *ClipsDatabase::ProcessLoop(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent)
{
  *returnType = ClipsNull;
  ClipsExpr *loopExpr = expr->Nth(1);
  ClipsType typ;
  if (loopExpr->Type() != ClipsList)
  {
    char buf[200];
    sprintf(buf, "In function %s, loop-for-count must use the variable form.", func->GetName());
    ClipsError(buf);
    return expr;
  }

  ClipsExpr *varExpr = loopExpr->Nth(0);
  ClipsExpr *fromExpr = loopExpr->Nth(1);
  ClipsExpr *toExpr = loopExpr->Nth(2);
  if (!varExpr || !fromExpr || !toExpr || varExpr->Type() != ClipsWord)
  {
    char buf[200];
    sprintf(buf, "In function %s, syntax error in loop-for-count.", func->GetName());
    ClipsError(buf);
    return expr;
  }

  char *name = varExpr->WordValue();
  if (name[0] == '?')
    name = name + 1;

  ClipsConstruct *var = func->locals.FindConstruct(name);
  if (!var)
    var = TheNameSpace.FindConstruct(name);
  if (!var)
  {
    func->locals.AddConstruct(name, CONSTRUCT_TYPE_LOCAL);
    var = func->locals.FindConstruct(name);
  }
  
  var->SetReturnType(ClipsInteger);
  
  ClipsExpr *statement = expr->Nth(2);
  if ((statement->Type() == ClipsWord) && (strcmp(statement->WordValue(), "do") == 0))
    statement = statement->GetNext();

  while (statement)
  {
    ProcessExpression(func, statement, &typ, indent);
    statement = statement->GetNext();
  }
  
  expr->SetReturnType(ClipsNull);
  return expr;
}

ClipsExpr *ClipsDatabase::ProcessSwitch(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent)
{
  *returnType = ClipsNull;
  ClipsExpr *switchExpr = expr->Nth(1);
  ClipsType typ;
  ProcessExpression(func, switchExpr, &typ, indent);
  
  ClipsExpr *caseList = expr->Nth(2);

  while (caseList)
  {
    ClipsExpr *caseExpr = caseList->Nth(0);
    ClipsExpr *compExpr = caseList->Nth(1);
    ClipsExpr *thenExpr = caseList->Nth(2);
    ClipsExpr *actionExpr = caseList->Nth(3);
    if (caseExpr && strcmp(caseExpr->WordValue(), "case") == 0)
    {
      if (compExpr)
        ProcessExpression(func, compExpr, &typ, indent);
    }
    else if (caseExpr && strcmp(caseExpr->WordValue(), "default") == 0)
      actionExpr = caseList->Nth(1);

    while (actionExpr)
    {
      ProcessExpression(func, actionExpr, &typ, indent);
      actionExpr = actionExpr->GetNext();
    }

    caseList = caseList->GetNext();
  }    
  
  expr->SetReturnType(ClipsNull);
  return expr;
}

ClipsExpr *ClipsDatabase::ProcessBind(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent)
{
  char buf[200];
  *returnType = ClipsUnknown;
  expr->SetReturnType(ClipsNull);
  ClipsExpr *nameExpr = expr->Nth(1);
  if (nameExpr->Type() != ClipsWord)
  {
    ClipsError("Expected a name following 'bind'");
    return expr;
  }
  char *name = nameExpr->WordValue();
  if (name[0] == '?')
    name = name + 1;

  ClipsConstruct *var = func->locals.FindConstruct(name);
  if (!var)
    var = TheNameSpace.FindConstruct(name);
  if (!var)
  {
    func->locals.AddConstruct(name, CONSTRUCT_TYPE_LOCAL);
    var = func->locals.FindConstruct(name);
  }
    
  ClipsType constructType = var->GetReturnType();

  ClipsExpr *valueExpr = expr->Nth(2);
  if (!valueExpr)
  {
    sprintf(buf, "Function %s: no value", func->GetName());
    ClipsError(buf);
    return expr;
  }
  ClipsExpr *boundExpr = ProcessExpression(func, valueExpr, returnType, indent);
  switch (*returnType)
  {
    case ClipsInteger:
    {
      if (constructType == ClipsReal || constructType == ClipsInteger)
      // Do nothing: we've got a sufficient type.
      {}
      else if (constructType == ClipsUnknown)
        var->returnType = ClipsInteger;
      else
      {
        sprintf(buf, "Function %s: ambiguous type for %s", func->GetName(), name);
        ClipsError(buf);
        return expr;
      }
      break;
    }
    case ClipsReal:
    {
      if (constructType == ClipsReal)
      // Do nothing: we've got a sufficient type.
      {}
      // (further) constain type from unknown or integer
      else if (constructType == ClipsUnknown || constructType == ClipsInteger)
        var->returnType = ClipsReal;
      else
      {
        sprintf(buf, "Function %s: ambiguous type for %s", func->GetName(), name);
        ClipsError(buf);
        return expr;
      }
      break;
    }
    case ClipsString:
    case ClipsWord:
    {
      if (constructType == ClipsUnknown)
        var->returnType = ClipsString;
      else if (constructType == ClipsString || constructType == ClipsStringClass)
      {}
      else
      {
        sprintf(buf, "Function %s: ambiguous type for %s", func->GetName(), name);
        ClipsError(buf);
        return expr;
      }
      break;
    }
    case ClipsStringClass:
    {
      if (constructType == ClipsUnknown)
        var->returnType = ClipsStringClass;
      else if (constructType == ClipsStringClass)
      {}
      else
      {
        sprintf(buf, "Function %s: ambiguous type for %s", func->GetName(), name);
        ClipsError(buf);
        return expr;
      }
      break;
    }
    case ClipsList:
    {
      if (constructType == ClipsList)
      // Do nothing: we've got a sufficient type.
      {}
      else if (constructType == ClipsUnknown)
        var->returnType = ClipsList;
      else
      {
        sprintf(buf, "Function %s: ambiguous type for %s", func->GetName(), name);
        ClipsError(buf);
        return expr;
      }
      break;
    }
    case ClipsNull:
    case ClipsUnknown:
    default:
    {
      // Do what??
      break;
    }
  }
  return boundExpr;
}

ClipsExpr *ClipsDatabase::ProcessEq(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent)
{
  *returnType = ClipsUnknown;
  ClipsType eqType = ClipsUnknown;
  ClipsExpr *currentExpr = expr->Nth(1);
  // Try to ascertain type
  while (currentExpr)
  {
    ClipsType tmpType = ClipsUnknown;
    ProcessExpression(func, currentExpr, &tmpType, indent);
    if (tmpType != ClipsUnknown)
      eqType = tmpType;
      
    currentExpr = currentExpr->GetNext();
  }

  currentExpr = expr->Nth(1);
  while (currentExpr)
  {
    currentExpr->SetReturnType(eqType);
    ResolveVariableType(func, currentExpr, eqType);
    currentExpr = currentExpr->GetNext();
  }
  
  expr->SetReturnType(ClipsNull);
  return expr;
}

ClipsExpr *ClipsDatabase::ProcessNeq(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent)
{
  *returnType = ClipsUnknown;
  ClipsType eqType = ClipsUnknown;
  ClipsExpr *currentExpr = expr->Nth(1);
  // Try to ascertain type
  while (currentExpr)
  {
    ClipsType tmpType = ClipsUnknown;
    ProcessExpression(func, currentExpr, &tmpType, indent);
    if (tmpType != ClipsUnknown)
      eqType = tmpType;
      
    currentExpr = currentExpr->GetNext();
  }

  currentExpr = expr->Nth(1);
  while (currentExpr)
  {
    currentExpr->SetReturnType(eqType);
    ResolveVariableType(func, currentExpr, eqType);
    currentExpr = currentExpr->GetNext();
  }
  
  expr->SetReturnType(ClipsNull);
  return expr;
}

/*
 * Try to resolve the type of this expression: if a variable name, we
 * can resolve its type. Otherwise, just set the expression's return type.
 *
 */
 
void ClipsDatabase::ResolveVariableType(ClipsConstruct *func, ClipsExpr *argExpr, ClipsType theType)
{
  if ((argExpr->Type() == ClipsWord) && (argExpr->WordValue()[0] == '?'))
  {
    char *argName = argExpr->WordValue() + 1;
    ClipsConstruct *argConstruct = func->locals.FindConstruct(argName);
    if (!argConstruct)
      argConstruct = TheNameSpace.FindConstruct(argName);

    // Can resolve this variable now!
    if (argConstruct && (argConstruct->returnType == ClipsUnknown))
      argConstruct->returnType = theType;
  }
  argExpr->SetReturnType(theType);
}

ClipsExpr *ClipsDatabase::ProcessUserFunction(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent)
{
  char buf[200];
  *returnType = ClipsUnknown;
  char *name = expr->Nth(0)->WordValue();

  ClipsFunctionRecord *rec = TheFunctionTable.FindFunction(name);
  if (!rec)
  {
    sprintf(buf, "Unknown function %s called from function %s", name,
            func->GetName());
    ClipsError(buf);
    return expr;
  }

  /*
   * Look at all the arguments of the function -- resolve any variable
   * types we can (particularly parameters).
   *
   */
  ClipsExpr *argExpr = expr->Nth(1);
  wxNode *argNode = rec->argList.First();
  
  while (argExpr)
  {
    ClipsArg *arg = NULL;
    if (argNode)
      arg = (ClipsArg *)argNode->Data();

    if (arg)
      ResolveVariableType(func, argExpr, arg->argType);
      
    // SHOULD PROCESS ARGUMENTS
    ClipsType argType;
    ProcessExpression(func, argExpr, &argType, indent);

    // Resolve the type of this expression.
    if (argExpr->GetReturnType() == ClipsUnknown)
      if (arg)
        argExpr->SetReturnType(arg->argType);
      else
        argExpr->SetReturnType(argType);

    argExpr = argExpr->GetNext();
    if (argNode) argNode = argNode->Next();
  }

  // Return type may be ClipsString when it will be coerced to
  // ClipsStringClass.
  
  *returnType = rec->GetReturnType();
  expr->SetReturnType(*returnType);
  return expr;
}

ClipsExpr *ClipsDatabase::ProcessClipsFunction(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent)
{
  *returnType = ClipsUnknown;
  char *name = expr->Nth(0)->WordValue();

  ClipsConstruct *clipsFunc = TheNameSpace.FindConstruct(name);

  /*
   * Look at all the arguments of the function -- resolve any variable
   * types we can (particularly parameters).
   *
   */
  ClipsExpr *argExpr = expr->Nth(1);
  wxNode *argNode = clipsFunc->locals.First();
  
  while (argExpr)
  {
    ClipsConstruct *arg = NULL;
    if (argNode)
      arg = (ClipsConstruct  *)argNode->Data();

    if (arg && arg->GetReturnType() != ClipsUnknown)
      ResolveVariableType(func, argExpr, arg->GetReturnType());
      
    // SHOULD PROCESS ARGUMENTS
    ClipsType argType;
    ProcessExpression(func, argExpr, &argType, indent);

    // Resolve the type of this expression.
    if (argExpr->GetReturnType() == ClipsUnknown)
      if (arg)
        argExpr->SetReturnType(arg->GetReturnType());
      else
        argExpr->SetReturnType(argType);

    argExpr = argExpr->GetNext();
    if (argNode) argNode = argNode->Next();
  }

  // Return type may be ClipsString when it will be coerced to
  // ClipsStringClass.
  
  *returnType = clipsFunc->GetReturnType();
  expr->SetReturnType(*returnType);
  return expr;
}

ClipsExpr *ClipsDatabase::ProcessStrcat(ClipsConstruct *func, ClipsExpr *expr,
                             ClipsType *returnType, int indent)
{
  ClipsExpr *operand = expr->Nth(1);
  while (operand)
  {
    ProcessExpression(func, operand, returnType, indent);
    if (operand->GetReturnType() == ClipsUnknown)
      operand->SetReturnType(ClipsStringClass);
    ResolveVariableType(func, operand, ClipsString);
    operand = operand->GetNext();
  }
  *returnType = ClipsStringClass;
  expr->SetReturnType(ClipsStringClass);
  return expr;
}

ClipsExpr *ClipsDatabase::ProcessMvAppend(ClipsConstruct *func, ClipsExpr *expr,
                             ClipsType *returnType, int indent)
{
  ClipsExpr *operand = expr->Nth(1);
  while (operand)
  {
    ProcessExpression(func, operand, returnType, indent);
    if (operand->GetReturnType() == ClipsUnknown)
      operand->SetReturnType(*returnType);
    operand = operand->GetNext();
  }
  *returnType = ClipsList;
  expr->SetReturnType(ClipsList);
  return expr;
}

ClipsExpr *ClipsDatabase::ProcessMember(ClipsConstruct *func, ClipsExpr *expr,
                             ClipsType *returnType, int indent)
{
  ClipsExpr *operand1 = expr->Nth(1);
  ClipsExpr *operand2 = expr->Nth(2);
  ClipsType typ;
  ProcessExpression(func, operand1, &typ, indent);
  ProcessExpression(func, operand2, &typ, indent);
  if (typ != ClipsUnknown && typ != ClipsList)
  {
    char buf[100];
    sprintf(buf, "In function %s, second argument to member must be a list.",
       func->GetName());
    ClipsError(buf);
  }
  operand2->SetReturnType(ClipsList);
  expr->SetReturnType(ClipsInteger);
  *returnType = ClipsInteger;
  ResolveVariableType(func, operand2, ClipsList);
  return expr;
}

ClipsExpr *ClipsDatabase::ProcessLength(ClipsConstruct *func, ClipsExpr *expr,
                             ClipsType *returnType, int indent)
{
  ClipsExpr *operand = expr->Nth(1);
  ClipsType typ;
  ProcessExpression(func, operand, &typ, indent);
  if (typ != ClipsUnknown && typ != ClipsString && typ != ClipsWord && typ != ClipsList)
  {
    char buf[100];
    sprintf(buf, "In function %s, argument to length must be a list or a string.",
       func->GetName());
    ClipsError(buf);
  }
  expr->SetReturnType(ClipsInteger);
  *returnType = ClipsInteger;
  return expr;
}

ClipsExpr *ClipsDatabase::ProcessFormat(ClipsConstruct *func, ClipsExpr *expr,
                             ClipsType *returnType, int indent)
{
  ClipsExpr *operand1 = expr->Nth(1);
  ClipsExpr *operand2 = expr->Nth(2);
  ClipsExpr *operand3 = expr->Nth(3);
  ClipsType typ;
  ProcessExpression(func, operand1, &typ, indent);
  ProcessExpression(func, operand2, &typ, indent);
  while (operand3)
  {
    ProcessExpression(func, operand3, &typ, indent);
    operand3 = operand3->GetNext();
  }
  expr->SetReturnType(ClipsString);
  *returnType = ClipsString;
  return expr;
}

ClipsExpr *ClipsDatabase::ProcessPrintout(ClipsConstruct *func, ClipsExpr *expr,
                             ClipsType *returnType, int indent)
{
  ClipsExpr *operand1 = expr->Nth(1);
  ClipsExpr *operand2 = expr->Nth(2);
  ClipsType typ;
  ProcessExpression(func, operand1, &typ, indent);
  while (operand2)
  {
    ProcessExpression(func, operand2, &typ, indent);
    operand2 = operand2->GetNext();
  }
  expr->SetReturnType(ClipsNull);
  *returnType = ClipsNull;
  return expr;
}

/*
ClipsExpr *ClipsDatabase::ProcessDeleteSubseq(ClipsConstruct *func, ClipsExpr *expr,
                             ClipsType *returnType, int indent)
{
  ClipsExpr *op = expr->Nth(0);
  char *opName = op->WordValue();
  ClipsExpr *mvExpr = expr->Nth(1);
  ClipsExpr *operand1 = expr->Nth(2);
  ClipsExpr *operand2 = expr->Nth(3);
  ClipsType typ;
  ProcessExpression(func, mvExpr, &typ, indent);
  if (typ != ClipsUnknown && typ != ClipsList)
  {
    char buf[100];
    sprintf(buf, "In function %s, first argument to %s must be a list.",
       opName, func->GetName());
    ClipsError(buf);
  }
  ProcessExpression(func, operand1, &typ, indent);
  if (typ != ClipsUnknown && typ != ClipsInteger)
  {
    char buf[100];
    sprintf(buf, "In function %s, second argument to %s must be an integer.",
       opName, func->GetName());
    ClipsError(buf);
  }
  ProcessExpression(func, operand2, &typ, indent);
  if (typ != ClipsUnknown && typ != ClipsInteger)
  {
    char buf[100];
    sprintf(buf, "In function %s, third argument to %s must be an integer.",
       opName, func->GetName());
    ClipsError(buf);
  }
  mvExpr->SetReturnType(ClipsList);
  operand1->SetReturnType(ClipsInteger);
  operand2->SetReturnType(ClipsInteger);
  expr->SetReturnType(ClipsList);
  *returnType = ClipsList;
  ResolveVariableType(func, mvExpr, ClipsList);
  ResolveVariableType(func, operand1, ClipsInteger);
  ResolveVariableType(func, operand2, ClipsInteger);
  return expr;
}
*/
			     
ClipsExpr *ClipsDatabase::ProcessNth(ClipsConstruct *func, ClipsExpr *expr,
                             ClipsType *returnType, int indent)
{
  ClipsExpr *listExpr = expr->Nth(2);
  ClipsExpr *operand = expr->Nth(1);
  ResolveVariableType(func, operand, ClipsInteger);

  ClipsType lType;
  ProcessExpression(func, listExpr, &lType, indent);
  if (lType != ClipsUnknown && lType != ClipsList)
  {
    char buf[100];
    sprintf(buf, "In function %s, second argument to nth must be a list expression.",
       func->GetName());
    ClipsError(buf);
  }

  ClipsType nType;
  ProcessExpression(func, operand, &nType, indent);

  if (nType != ClipsUnknown && nType != ClipsInteger && nType != ClipsReal)
  {
    char buf[100];
    sprintf(buf, "In function %s, first argument to nth must be an integer expression.",
       func->GetName());
    ClipsError(buf);
  }
  operand->SetReturnType(ClipsInteger);

  *returnType = ClipsUnknown;
  expr->SetReturnType(ClipsUnknown);
  return expr;
}

ClipsExpr *ClipsDatabase::ProcessArithmetic(ClipsConstruct *func, ClipsExpr *expr,
                             ClipsType *returnType, int indent)
{
  ClipsExpr *operand = expr->Nth(1);
  // By default, integer; but see if we can make it real.
  ClipsType rType = ClipsInteger;
  while (operand)
  {
    ClipsType tmpType = ClipsUnknown;
    ProcessExpression(func, operand, &tmpType, indent);
    if (tmpType == ClipsReal)
      rType = ClipsReal;
    operand = operand->GetNext();
  }
  *returnType = rType;
  expr->SetReturnType(rType);

  // See if we can resolve any variable types
  operand = expr->Nth(1);
  while (operand)
  {
    ResolveVariableType(func, operand, rType);
    operand = operand->GetNext();
  }
  return expr;
}

ClipsExpr *ClipsDatabase::ProcessComparison(ClipsConstruct *func, ClipsExpr *expr,
                             ClipsType *returnType, int indent)
{
  ClipsExpr *operand = expr->Nth(1);
  // By default, integer; but see if we can make it real.
  ClipsType rType = ClipsInteger;
  while (operand)
  {
    ClipsType tmpType = ClipsUnknown;
    ProcessExpression(func, operand, &tmpType, indent);
    if (tmpType == ClipsReal)
      rType = ClipsReal;
    operand = operand->GetNext();
  }
  *returnType = ClipsInteger;
  expr->SetReturnType(ClipsInteger);

  // See if we can resolve any variable types
  operand = expr->Nth(1);
  while (operand)
  {
    ResolveVariableType(func, operand, rType);
    operand = operand->GetNext();
  }
  return expr;
}

ClipsExpr *ClipsDatabase::ProcessPredicate(ClipsConstruct *func, ClipsExpr *expr,
                             ClipsType *returnType, int indent)
{
  ClipsExpr *pred = expr->Nth(0);
  char *predName = pred->WordValue();
  ClipsExpr *operand = expr->Nth(1);

  expr->SetReturnType(ClipsInteger);
  *returnType = ClipsInteger;
  return expr;
}

/*
 *
 * GENERATION
 *
 */
                               
Bool ClipsDatabase::GenerateExpression(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent, ostream& stream)
{
  switch (expr->Type())
  {
    case ClipsList:
    {
      if (expr->Nth(0) == NULL)
        return TRUE;
      ClipsExpr *first = expr->Nth(0);
      if (first->Type() != ClipsWord)
        return TRUE;
      // Ok, now we have a function, either built-in or user-defined.
      char *name = first->WordValue();
      if (strcmp(name, "if") == 0)
        return GenerateIf(func, expr, returnType, indent, stream);
      else if (strcmp(name, "declare") == 0)
      // Ignore
      {}
      else if (strcmp(name, "return") == 0)
        return GenerateReturn(func, expr, returnType, indent, stream);
      else if (strcmp(name, "while") == 0)
        return GenerateWhile(func, expr, returnType, indent, stream);
      else if (strcmp(name, "progn") == 0 || strcmp(name, "progn$") == 0)
        return GenerateProgn(func, expr, returnType, indent, stream);
      else if (strcmp(name, "switch") == 0)
        return GenerateSwitch(func, expr, returnType, indent, stream);
      else if (strcmp(name, "loop-for-count") == 0)
        return GenerateLoop(func, expr, returnType, indent, stream);
      else if (strcmp(name, "break") == 0)
        return GenerateBreak(func, expr, returnType, indent, stream);
      else if (strcmp(name, "bind") == 0)
        return GenerateBind(func, expr, returnType, indent, stream);
      else if (strcmp(name, "eq") == 0)
        return GenerateEq(func, expr, returnType, indent, stream);
      else if (strcmp(name, "neq") == 0)
        return GenerateNeq(func, expr, returnType, indent, stream);
      else if (strcmp(name, "str-cat") == 0 || strcmp(name, "sym-cat") == 0)
        return GenerateStrcat(func, expr, returnType, indent, stream);
      else if (strcmp(name, "upcase") == 0)
        return GenerateUpcase(func, expr, returnType, indent, stream);
      else if (strcmp(name, "lowcase") == 0)
        return GenerateLowcase(func, expr, returnType, indent, stream);
      else if (strcmp(name, "substring") == 0)
        return GenerateSubstring(func, expr, returnType, indent, stream);
      else if (strcmp(name, "mv-append") == 0 || strcmp(name, "create$") == 0 || strcmp(name, "create") == 0)
        return GenerateMvAppend(func, expr, returnType, indent, stream);
      else if (strcmp(name, "member") == 0 || strcmp(name, "member$") == 0)
        return GenerateMember(func, expr, returnType, indent, stream);
      else if (strcmp(name, "length") == 0 || strcmp(name, "length$") == 0)
        return GenerateLength(func, expr, returnType, indent, stream);
      else if (strcmp(name, "format") == 0)
        return GenerateFormat(func, expr, returnType, indent, stream);
      else if (strcmp(name, "printout") == 0)
        return GeneratePrintout(func, expr, returnType, indent, stream);
      else if (strcmp(name, "read") == 0)
        return GenerateRead(func, expr, returnType, indent, stream);
      else if (strcmp(name, "readline") == 0)
        return GenerateReadline(func, expr, returnType, indent, stream);
      else if (strcmp(name, "open") == 0)
        return GenerateOpen(func, expr, returnType, indent, stream);
      else if (strcmp(name, "close") == 0)
        return GenerateClose(func, expr, returnType, indent, stream);
      else if (strcmp(name, "delete") == 0 || strcmp(name, "delete$") == 0 ||
               strcmp(name, "subseq") == 0 || strcmp(name, "subseq$") == 0)
        return GenerateDeleteSubseq(func, expr, returnType, indent, stream);
      else if (strcmp(name, "explode") == 0 || strcmp(name, "explode$") == 0)
        return GenerateExplode(func, expr, returnType, indent, stream);
      else if (strcmp(name, "implode") == 0 || strcmp(name, "implode$") == 0)
        return GenerateImplode(func, expr, returnType, indent, stream);
      else if (strcmp(name, "insert") == 0 || strcmp(name, "insert$") == 0)
        return GenerateInsert(func, expr, returnType, indent, stream);
      else if (strcmp(name, "replace") == 0 || strcmp(name, "replace$") == 0)
        return GenerateReplace(func, expr, returnType, indent, stream);
      else if (strcmp(name, "nth") == 0 || strcmp(name, "nth$") == 0)
        return GenerateNth(func, expr, returnType, indent, stream);
      else if (strcmp(name, "first") == 0 || strcmp(name, "first$") == 0 ||
               strcmp(name, "rest") == 0 || strcmp(name, "rest$") == 0)
        return GenerateFirstRest(func, expr, returnType, indent, stream);
      else if (strcmp(name, "not") == 0 || strcmp(name, "or") == 0 ||
               strcmp(name, "and") == 0)
        return GenerateLogicalOp(func, expr, returnType, indent, stream);
      else if (strcmp(name, "+") == 0 || strcmp(name, "-") == 0 ||
               strcmp(name, "*") == 0 || strcmp(name, "/") == 0)
        return GenerateArithmetic(func, expr, returnType, indent, stream);
      else if (strcmp(name, "min") == 0 || strcmp(name, "max") == 0)
        return GenerateMinMax(func, expr, returnType, indent, stream);
      else if (strcmp(name, "<") == 0 || strcmp(name, ">") == 0 ||
               strcmp(name, "<=") == 0 || strcmp(name, ">=") == 0 ||
               strcmp(name, "=") == 0 ||  strcmp(name, "<>") == 0)
        return GenerateComparison(func, expr, returnType, indent, stream);
      else if (strcmp(name, "numberp") == 0 || strcmp(name, "floatp") == 0 ||
               strcmp(name, "integerp") == 0 || strcmp(name, "lexemep") == 0 ||
               strcmp(name, "stringp") == 0 || strcmp(name, "symbolp") == 0 ||
               strcmp(name, "evenp") == 0 || strcmp(name, "oddp") == 0 ||
               strcmp(name, "multifieldp") == 0)
        return GeneratePredicate(func, expr, returnType, indent, stream);
      else
      {
        ClipsConstruct *clipsFunc = TheNameSpace.FindConstruct(name);
        if (clipsFunc && (clipsFunc->GetType() == CONSTRUCT_TYPE_FUNCTION))
          return GenerateClipsFunction(func, expr, returnType, indent, stream);
        else
          return GenerateUserFunction(func, expr, returnType, indent, stream);
      }

      break;
    }
    case ClipsWord:
    {
      // Check for variable
      char *name = expr->WordValue();
      if (name[0] == '?')
      {
        char buf[300];
        name = name + 1;
        ClipsConstruct *construct = func->locals.FindConstruct(name);
        if (!construct)
          construct = TheNameSpace.FindConstruct(name);
        if (!construct)
        {
          sprintf(buf, "Undeclared variable %s in function %s", name, func->GetName());
          ClipsError(buf);
//          return FALSE;
        }
        else
        {
          if (construct->GetReturnType() == ClipsNull)
            Indent(stream, indent);
        }
        strcpy(buf, name);
        ClipsNameToC(buf);

        stream << buf;
      }
      else
      {
        if (strcmp(name, "TRUE") == 0 || strcmp(name, "FALSE") == 0)
        {
          stream << name << " ";
        }
        else
        {
          stream << "wxString(\"" << name << "\")";
        }
      }
      return TRUE;
      break;
    }
    case ClipsInteger:
    {
      stream << expr->IntegerValue();
      return TRUE;
      break;
    }
    case ClipsReal:
    {
      stream << "((float)" << expr->RealValue() << ")";
      return TRUE;
      break;
    }
    case ClipsString:
    {
      stream << "wxString(\"" << expr->StringValue() << "\")";
      return TRUE;
      break;
    }
    case ClipsNull:
    case ClipsUnknown:
    default:
    {
      return TRUE;
      break;
    }
  }
  return TRUE;
}

Bool ClipsDatabase::GenerateIf(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent, ostream& stream)
{
  *returnType = ClipsUnknown;
  
  ClipsExpr *thing = expr->GetFirst();
  // Ignore 'if'
  thing = thing->GetNext();

  Indent(stream, indent);
  stream << "if (";

  // Next one is the test
  ClipsType typ;
  if (!GenerateExpression(func, thing, &typ, indent, stream))
    return FALSE;

  stream << ")\n";

  // Next one is 'then'
  thing = thing->GetNext();
  if ((thing->Type() != ClipsWord) || (strcmp(thing->WordValue(), "then") != 0))
  {
    ClipsError("Expected a 'then'");
  }

  Indent(stream, indent);
  stream << "{\n";

  // Next one is the 'then' expr
  thing = thing->GetNext();
  ClipsExpr *thenExpr = NULL;
  while (thing && !((thing->Type() == ClipsWord && strcmp(thing->WordValue(), "else") == 0)))
  {
//    Indent(stream, indent);
    // This expression is effectively void within a 'then'.
    thing->SetReturnType(ClipsNull);
    GenerateExpression(func, thing, returnType, indent+2, stream);
    stream << ";\n";
    thing = thing->GetNext();
  }

  Indent(stream, indent);
  stream << "}\n";

  // Is there an 'else'?  
  if (thing)
  {
    if ((thing->Type() != ClipsWord) || (strcmp(thing->WordValue(), "else") != 0))
    {
      ClipsError("Expected an 'else'");
      return FALSE;
    }

    // Yes, so...
    Indent(stream, indent);
    stream << "else\n";
    Indent(stream, indent);
    stream << "{\n";
    thing = thing->GetNext();

    while (thing)
    {
//      Indent(stream, indent);
      // This expression is effectively void within an 'else'.
      thing->SetReturnType(ClipsNull);
      GenerateExpression(func, thing, returnType, indent+2, stream);
      stream << ";\n";
      thing = thing->GetNext();
    }
    Indent(stream, indent);
    stream << "}\n";
  }
  return TRUE;
}

Bool ClipsDatabase::GenerateWhile(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent, ostream& stream)
{
  ClipsExpr *testExpr = expr->Nth(1);
  Indent(stream, indent);
  stream << "while (";
  GenerateExpression(func, testExpr, returnType, indent, stream);
  stream << ")\n";
  Indent(stream, indent);
  stream << "{\n";
  ClipsExpr *statement = testExpr->GetNext();
  if (statement && (statement->Type() == ClipsWord) &&
       (strcmp(statement->WordValue(), "do") == 0))
    statement = statement->GetNext();
  while (statement)
  {
//    Indent(stream, indent);
    // This expression is effectively void within a 'while'.
    statement->SetReturnType(ClipsNull);
    GenerateExpression(func, statement, returnType, indent+2, stream);
    stream << ";\n";
    statement = statement->GetNext();
  }
  Indent(stream, indent);
  stream << "}\n";
  return TRUE;
}

Bool ClipsDatabase::GenerateProgn(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent, ostream& stream)
{
  Indent(stream, indent);
  stream << "{\n";
  ClipsExpr *statement = expr->Nth(1);
  while (statement)
  {
    // This expression is effectively void within a 'progn'.
    statement->SetReturnType(ClipsNull);
    GenerateExpression(func, statement, returnType, indent+2, stream);
    stream << ";\n";
    statement = statement->GetNext();
  }
  Indent(stream, indent);
  stream << "}\n";
  return TRUE;
}

Bool ClipsDatabase::GenerateSwitch(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent, ostream& stream)
{
  *returnType = ClipsNull;
  ClipsExpr *switchExpr = expr->Nth(1);
  ClipsType typ;
  
  ClipsExpr *caseList = expr->Nth(2);

  Indent(stream, indent);
  while (caseList)
  {
    ClipsExpr *caseExpr = caseList->Nth(0);
    ClipsExpr *compExpr = caseList->Nth(1);
    ClipsExpr *thenExpr = caseList->Nth(2);
    ClipsExpr *actionExpr = caseList->Nth(3);
    if (caseExpr && strcmp(caseExpr->WordValue(), "case") == 0)
    {
      stream << "if (";

      GenerateExpression(func, switchExpr, &typ, indent, stream);
      stream << " == ";
      if (compExpr)
        GenerateExpression(func, compExpr, &typ, indent, stream);
      stream << ")";
    }
    else if (caseExpr && strcmp(caseExpr->WordValue(), "default") == 0)
      actionExpr = caseList->Nth(1);

    stream << "\n";
    Indent(stream, indent);
    stream << "{\n";
    while (actionExpr)
    {
      actionExpr->SetReturnType(ClipsNull);
      GenerateExpression(func, actionExpr, &typ, indent+2, stream);
      stream << ";\n";
      actionExpr = actionExpr->GetNext();
    }
    Indent(stream, indent);
    stream << "}\n";

    caseList = caseList->GetNext();
    if (caseList)
    {
      Indent(stream, indent);
      stream << "else ";
    }
  }    
  
  return TRUE;
}

Bool ClipsDatabase::GenerateLoop(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent, ostream& stream)
{
  *returnType = ClipsNull;
  ClipsExpr *loopExpr = expr->Nth(1);
  ClipsType typ;
  if (loopExpr->Type() != ClipsList)
  {
    return FALSE;
  }

  ClipsExpr *varExpr = loopExpr->Nth(0);
  ClipsExpr *fromExpr = loopExpr->Nth(1);
  ClipsExpr *toExpr = loopExpr->Nth(2);
  if (!varExpr || !fromExpr || !toExpr || varExpr->Type() != ClipsWord)
  {
    return FALSE;
  }

  char *name = varExpr->WordValue();
  if (name[0] == '?')
    name = name + 1;

  ClipsConstruct *var = func->locals.FindConstruct(name);
  if (!var) return FALSE;
    
  if (expr->GetReturnType() == ClipsNull)
    Indent(stream, indent);

  char cname[100];
  strcpy(cname, var->GetCName());
  
  stream << "for (" << cname << " = ";
  GenerateExpression(func, fromExpr, &typ, indent, stream);
  stream << "; " << cname << " <= ";
  GenerateExpression(func, toExpr, &typ, indent, stream);
  stream << "; " << cname << " ++)\n";
  Indent(stream, indent);
  stream << "{\n";

  ClipsExpr *statement = expr->Nth(2);
  if ((statement->Type() == ClipsWord) && (strcmp(statement->WordValue(), "do") == 0))
    statement = statement->GetNext();

  while (statement)
  {
    statement->SetReturnType(ClipsNull);
    GenerateExpression(func, statement, &typ, indent+2, stream);
    stream << ";\n";
    statement = statement->GetNext();
  }

  Indent(stream, indent);
  stream << "}\n";
  
  return TRUE;
}

Bool ClipsDatabase::GenerateBreak(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent, ostream& stream)
{
  Indent(stream, indent);
  stream << "break";
  return TRUE;
}

Bool ClipsDatabase::GenerateBind(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent, ostream& stream)
{
  ClipsExpr *nameExpr = expr->Nth(1);
  Indent(stream, indent);
  if (nameExpr)
  {
    char *name = nameExpr->WordValue();
    name += 1;
    char buf[200];
    strcpy(buf, name);
    ClipsNameToC(buf);
    stream << buf << " = ";
    ClipsExpr *valExpr = nameExpr->GetNext();
    if (valExpr)
    {
      Bool succ = GenerateExpression(func, valExpr, returnType, indent, stream);
      if (succ)
      {
        return TRUE;
      }
    }
  }
  return FALSE;
}

Bool ClipsDatabase::GenerateEq(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent, ostream& stream)
{
  ClipsExpr *first = expr->Nth(1);
  ClipsExpr *second = expr->Nth(2);
  if (!first || !second)
    return FALSE;
  stream << "(";
  GenerateExpression(func, first, returnType, indent, stream);
  stream << " == ";
  GenerateExpression(func, second, returnType, indent, stream);
  stream << ")";
  return TRUE;
}

Bool ClipsDatabase::GenerateNeq(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent, ostream& stream)
{
  ClipsExpr *first = expr->Nth(1);
  ClipsExpr *second = expr->Nth(2);
  if (!first || !second)
    return FALSE;
  stream << "(";
  GenerateExpression(func, first, returnType, indent, stream);
  stream << " != ";
  GenerateExpression(func, second, returnType, indent, stream);
  stream << ")";
  return TRUE;
}

Bool ClipsDatabase::GenerateLogicalOp(ClipsConstruct *func, ClipsExpr *expr,
                             ClipsType *returnType, int indent, ostream& stream)
{
  stream << "(";
  ClipsExpr *first = expr->GetFirst();
  char *op = first->WordValue();
  ClipsExpr *operand = first->GetNext();
  if (strcmp(op, "not") == 0)
  {
    stream << "! ";
    GenerateExpression(func, operand, returnType, indent, stream);
  }
  else
  {
    while (operand)
    {
      GenerateExpression(func, operand, returnType, indent, stream);
      operand = operand->GetNext();
      if (operand)
      {
        if (strcmp(op, "and") == 0)
          stream << " && ";
        else if (strcmp(op, "or") == 0)
          stream << " || ";
      }
    }
  }
  stream << ")";
  return TRUE;
}

Bool ClipsDatabase::GenerateArithmetic(ClipsConstruct *func, ClipsExpr *expr,
                             ClipsType *returnType, int indent, ostream& stream)
{
  stream << "(";
  ClipsExpr *first = expr->GetFirst();
  char *op = first->WordValue();
  ClipsExpr *operand = first->GetNext();
/*
  if (strcmp(op, "-") == 0)
  {
    stream << "! ";
    GenerateExpression(func, operand, returnType, indent, stream);
  }
  else
*/
  {
    while (operand)
    {
      GenerateExpression(func, operand, returnType, indent, stream);
      operand = operand->GetNext();
      if (operand)
      {
        stream << op;
      }
    }
  }
  stream << ")";
  return TRUE;
}    

Bool ClipsDatabase::GenerateMinMax(ClipsConstruct *func, ClipsExpr *expr,
                             ClipsType *returnType, int indent, ostream& stream)
{
  stream << "(";
  ClipsExpr *first = expr->GetFirst();
  char *op = first->WordValue();
  ClipsExpr *operand = first->GetNext();

  if (strcmp(op, "min") == 0)
    stream << "ClipsMin(MakeList(";
  else
    stream << "ClipsMax(MakeList(";

  while (operand)
  {
    GenerateExpression(func, operand, returnType, indent, stream);
    operand = operand->GetNext();
    stream << ", ";
  }
  stream << "NULL))";
  return TRUE;
}    

Bool ClipsDatabase::GenerateComparison(ClipsConstruct *func, ClipsExpr *expr,
                             ClipsType *returnType, int indent, ostream& stream)
{
  stream << "(";
  ClipsExpr *first = expr->GetFirst();
  char *op = first->WordValue();
  ClipsExpr *operand = first->GetNext();
  while (operand)
  {
    GenerateExpression(func, operand, returnType, indent, stream);
    operand = operand->GetNext();
    if (operand)
    {
      if (strcmp(op, "=") == 0)
        stream << "==";
      else if (strcmp(op, "<>") == 0)
        stream << "!=";
      else
        stream << op;
    }
  }
  stream << ")";
  return TRUE;
}    

Bool ClipsDatabase::GenerateUserFunction(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent, ostream& stream)
{
  char buf[200];
  ClipsExpr *currentExpr = expr->Nth(0);
  if (currentExpr->Type() != ClipsWord)
  {
    char buf[200];
    sprintf(buf, "In function %s, expected a function.", func->GetName());
    ClipsError(buf);
    return FALSE;
  }

  char *name = currentExpr->WordValue();

  ClipsFunctionRecord *clipsFunc = TheFunctionTable.FindFunction(name);
  if (!clipsFunc)
  {
    sprintf(buf, "Function %s not known.", name);
    ClipsError(buf);
    return FALSE;
  }

  // A void function should be indented.
  if (expr->GetReturnType() == ClipsNull)
    Indent(stream, indent);

  Bool coerceToStringClass = FALSE;
  if (clipsFunc &&
   (clipsFunc->GetReturnType() == ClipsString || clipsFunc->GetReturnType() == ClipsWord))
    coerceToStringClass = TRUE;

  if (coerceToStringClass)
    stream << "wxString(";
    
  char cName[100];
  if (clipsFunc)
    strcpy(cName, clipsFunc->GetCName());
  else
  {
    strcpy(cName, name);
    ClipsNameToC(cName);
  }

  stream << cName << "(";
  currentExpr = currentExpr->GetNext();
  wxNode *argNode = clipsFunc->argList.First();
  while (currentExpr)
  {
    Bool isSimpleString = FALSE;
    ClipsArg *arg = NULL;
    if (argNode)
    {
      arg = (ClipsArg *)argNode->Data();
      if (arg->argType == ClipsString)
        isSimpleString = TRUE;
    }
    else
    {
      sprintf(buf, "In function %s, too many arguments for function %s.",
           func->GetName(), clipsFunc->GetClipsName());
      ClipsError(buf);
    }

    // Optimize: if we have a literal string or word,
    // don't generate the expression with a wxString wrapper, just pass
    // the literal.
    if (isSimpleString && ((currentExpr->Type() == ClipsString) ||
                           (currentExpr->Type() == ClipsWord &&
                               (currentExpr->WordValue()[0] != '?') &&
                               (strcmp(currentExpr->WordValue(), "TRUE") != 0) &&
                               (strcmp(currentExpr->WordValue(), "t") != 0) &&
                               (strcmp(currentExpr->WordValue(), "FALSE") != 0) &&
                               (strcmp(currentExpr->WordValue(), "nil") != 0))))
    {
      stream << "\"" << currentExpr->StringValue() << "\"";
    }
    else if (arg && (arg->argType == ClipsFunction))
    {
      if (generateMode == MODE_HARDY)
        stream << "(HardyFunction)";
      else
        stream << "(wxClipsFunction)";
      char *fName = NULL;
      if (currentExpr->Type() == ClipsString)
        fName = currentExpr->StringValue();
      else if (currentExpr->Type() == ClipsWord)
        fName = currentExpr->WordValue();
      if (fName[0] == '?')  // Variables not allowed for now
        fName = NULL;
      if (fName)
      {
        if ((strlen(fName) == 0) || strcmp(fName, "nil") == 0)
          stream << "NULL";
        else
        {
          char fBuf[100];
          strcpy(fBuf, fName);
          ClipsNameToC(fBuf);
          stream << fBuf;
        }
      }
      else
      {
        stream << "NULL";
        sprintf(buf, "In function %s, function argument to %s must be a string or word.",
         func->GetName(), clipsFunc->GetClipsName());
        ClipsError(buf);
      }
    }
    else
    {
      GenerateExpression(func, currentExpr, returnType, indent, stream);
      if (isSimpleString)
        stream << ".GetData()";
    }
    
    currentExpr = currentExpr->GetNext();

    if (argNode)
      argNode = argNode->Next();

    if (currentExpr)
      stream << ", ";
  }
  stream << ")";

  if (coerceToStringClass)
    stream << ")";

  return TRUE;
}

Bool ClipsDatabase::GenerateClipsFunction(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent, ostream& stream)
{
  char buf[200];
  ClipsExpr *currentExpr = expr->Nth(0);
  if (currentExpr->Type() != ClipsWord)
  {
    char buf[200];
    sprintf(buf, "In function %s, expected a function.", func->GetName());
    ClipsError(buf);
    return FALSE;
  }

  char *name = currentExpr->WordValue();

  ClipsConstruct *clipsFunc = TheNameSpace.FindConstruct(name);
  if (!clipsFunc)
  {
    sprintf(buf, "Function %s not known.", name);
    ClipsError(buf);
    return FALSE;
  }

  // A void function should be indented.
  if (expr->GetReturnType() == ClipsNull)
    Indent(stream, indent);

  stream << clipsFunc->GetCName() << "(";
  currentExpr = currentExpr->GetNext();
  wxNode *argNode = clipsFunc->locals.First();
  while (currentExpr)
  {
    Bool isSimpleString = FALSE;
    ClipsConstruct *arg = NULL;
    if (argNode)
    {
      arg = (ClipsConstruct *)argNode->Data();
    }
    else
    {
      sprintf(buf, "In function %s, too many arguments for function %s.",
           func->GetName(), clipsFunc->GetName());
      ClipsError(buf);
    }
    GenerateExpression(func, currentExpr, returnType, indent, stream);
    
    currentExpr = currentExpr->GetNext();

    if (argNode)
      argNode = argNode->Next();

    if (currentExpr)
      stream << ", ";
  }
  stream << ")";

  return TRUE;
}

Bool ClipsDatabase::GenerateReturn(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent, ostream& stream)
{
  Indent(stream, indent);
  stream << "return ";
  ClipsExpr *valExpr = expr->Nth(1);
  if (valExpr)
  {
    if (!GenerateExpression(func, valExpr, returnType, indent, stream))
      return FALSE;
  }
  return TRUE;
}

Bool ClipsDatabase::GenerateFunction(ClipsConstruct *construct, ostream& stream)
{
  ClipsExpr *expr = construct->GetValue();
  ClipsExpr *funcName = expr->Nth(1);
  ClipsExpr *funcArgs = expr->Nth(2);
  char *originalName = funcName->WordValue();
  char buf[100];
  
  // Generate the function prototype
  switch (construct->GetReturnType())
  {
    case ClipsReal:
    {
      stream << "float ";
      break;
    }
    case ClipsInteger:
    {
      stream << "long ";
      break;
    }
    case ClipsString:
    case ClipsStringClass:
    case ClipsWord:
    {
      stream << "wxString ";
      break;
    }
    case ClipsList:
    {
      stream << "ClipsExpr ";
      break;
    }
    case ClipsNull:
    {
      stream << "void ";
      break;
    }
    default:
    {
      stream << "void ";
      sprintf(buf, "Could not resolve return type for function %s", originalName);
      ClipsError(buf);
      break;
    }
  }

  // May be a class member
  if (construct->GetUserType())
  {
    strcpy(buf, construct->GetUserType());
    ClipsNameToC(buf);
    stream << buf << "::";
  }
    
  strcpy(buf, originalName);
  ClipsNameToC(buf);
  stream << buf << "(";

  if (funcArgs->Nth(0) == NULL)
    stream << "void";
    
  ClipsExpr *arg = funcArgs->GetFirst();
  while (arg)
  {
    char *val = arg->WordValue();
    char *argName = val;
    if (val[0] == '?') argName = (char *)(val + 1);
    
    ClipsConstruct *argConstruct = construct->locals.FindConstruct(argName);
    if (argConstruct)
    {
       switch (argConstruct->GetReturnType())
       {
         case ClipsReal:
         {
          stream << "float ";
          break;
        }
        case ClipsInteger:
        {
          stream << "long ";
          break;
        }
        case ClipsString:
        case ClipsWord:
        {
          stream << "wxString ";
          break;
        }
        case ClipsList:
        {
          stream << "ClipsExpr ";
          break;
        }
        case ClipsUnknown:
        default:
        {
          stream << "long ";
          char buf2[200];
          sprintf(buf2, "In function '%s', could not resolve type for argument '%s': use 'declare'.",
               originalName, argName);
          ClipsWarning(buf2);
          break;
        }
      }
      strcpy(buf, argName);
      ClipsNameToC(buf);
      stream << buf;
    }
    arg = arg->GetNext();
    if (arg) stream << ", ";
  }
  stream << ")\n";
  stream << "{\n";

  // Generate locals here
  wxNode *node = construct->locals.First();
  while (node)
  {
    ClipsConstruct *localConstruct = (ClipsConstruct *)node->Data();
    if (localConstruct->GetType() == CONSTRUCT_TYPE_LOCAL)
    {
    stream << "  ";
    switch (localConstruct->GetReturnType())
    {
      case ClipsReal:
      {
        stream << "float ";
        break;
      }
      case ClipsInteger:
      {
        stream << "long ";
        break;
      }
      case ClipsString:
      case ClipsStringClass:
      case ClipsWord:
      {
        stream << "wxString ";
        break;
      }
      case ClipsList:
      {
        stream << "ClipsExpr ";
        break;
      }
      case ClipsUnknown:
      default:
      {
        stream << "long ";
        char buf2[200];
        sprintf(buf2, "In function '%s', could not resolve type for local '%s': use 'declare'.",
             originalName, localConstruct->GetName());
        ClipsWarning(buf2);
        break;
      }
    }
    strcpy(buf, localConstruct->GetName());
    ClipsNameToC(buf);
    stream << buf << ";\n";
    }
    node = node->Next();
  }

  // Generate body here
  ClipsExpr *bodyExpr = expr->Nth(3);
  while (bodyExpr)
  {
    ClipsType retType;
    // This expression is effectively void within a function.
    bodyExpr->SetReturnType(ClipsNull);
    GenerateExpression(construct, bodyExpr, &retType, 2, stream);
    stream << ";\n";
    bodyExpr = bodyExpr->GetNext();
  }

  // Generate a dummy return statement to satisfy the compiler
  switch (construct->GetReturnType())
  {
    case ClipsReal:
    {
      stream << "  return 0.0;\n";
      break;
    }
    case ClipsInteger:
    {
      stream << "  return 0;\n";
      break;
    }
    case ClipsString:
    case ClipsWord:
    {
      stream << "  return wxString("");\n";
      break;
    }
    case ClipsList:
    {
      stream << "  return ClipsExpr(ClipsUnknown);\n";
      break;
    }
    default:
    {
      break;
    }
  }

  stream << "}\n\n";
  return TRUE;
}

Bool ClipsDatabase::GenerateGlobal(ClipsConstruct *construct, ostream& stream)
{
  ClipsExpr *expr = construct->GetValue();
  
  ClipsExpr *globName = expr->Nth(1);
  ClipsExpr *globVal = expr->Nth(3);

  char buf[100];
  strcpy(buf, globName->WordValue());
  ClipsNameToC(buf);
  char *name = buf+1;
  switch (construct->GetReturnType())
  {
    case ClipsReal:
    {
      stream << "float " << name << " = ";
      break;
    }
    case ClipsInteger:
    {
      stream << "long " << name << " = ";
      break;
    }
    case ClipsString:
    case ClipsWord:
    {
      stream << "wxString " << name << " = ";
      break;
    }
    case ClipsList:
    {
      stream << "ClipsExpr " << name << " = ";
      break;
    }
    default:
    {
      stream << "<unknown global>";
      break;
    }
  }
  ClipsType retType;
  GenerateExpression(construct, globVal, &retType, 2, stream);

  stream << ";\n";

  return TRUE;
}

Bool ClipsDatabase::GenerateStatement(ClipsConstruct *construct, ostream& stream)
{
  ClipsType retType = ClipsNull;
  construct->GetValue()->SetReturnType(ClipsNull);
  GenerateExpression(construct, construct->GetValue(), &retType, 2, stream);
  stream << ";\n";
  return TRUE;
}

Bool ClipsDatabase::GeneratePrototypes(ostream& stream)
{
  stream << "/*\n * Function prototypes\n */\n\n";
  wxNode *node = ConstructList.First();
  while (node)
  {
    ClipsConstruct *construct = (ClipsConstruct *)node->Data();
    if (construct->GetType() == CONSTRUCT_TYPE_FUNCTION)
    {
      ClipsExpr *expr = construct->GetValue();
      ClipsExpr *funcName = expr->Nth(1);
      ClipsExpr *funcArgs = expr->Nth(2);
      char *originalName = funcName->WordValue();
      char buf[100];
  
      // Generate the function prototype
      switch (construct->GetReturnType())
      {
        case ClipsReal:
        {
          stream << "float ";
          break;
        }
        case ClipsInteger:
        {
          stream << "long ";
          break;
        }
        case ClipsString:
        case ClipsStringClass:
        case ClipsWord:
        {
          stream << "wxString ";
          break;
        }
        case ClipsList:
        {
          stream << "ClipsExpr ";
          break;
        }
        case ClipsNull:
        {
          stream << "void ";
          break;
        }
        default:
        {
          stream << "void ";
          break;
        }
      }
      strcpy(buf, originalName);
      ClipsNameToC(buf);
      stream << buf << "(";

      if (funcArgs->Nth(0) == NULL)
        stream << "void";
    
      ClipsExpr *arg = funcArgs->GetFirst();
      while (arg)
      {
        char *val = arg->WordValue();
        char *argName = val;
        if (val[0] == '?') argName = (char *)(val + 1);
    
        ClipsConstruct *argConstruct = construct->locals.FindConstruct(argName);
        if (argConstruct)
        {
          switch (argConstruct->GetReturnType())
          {
            case ClipsReal:
            {
              stream << "float ";
              break;
            }
            case ClipsInteger:
            {
              stream << "long ";
              break;
            }
            case ClipsString:
            case ClipsWord:
            {
              stream << "wxString ";
              break;
            }
            case ClipsList:
            {
              stream << "ClipsExpr ";
              break;
            }
            case ClipsUnknown:
            default:
            {
              stream << "long ";
              break;
            }
          }
          strcpy(buf, argName);
          ClipsNameToC(buf);
          stream << buf;
        }
        arg = arg->GetNext();
        if (arg) stream << ", ";
      }
      stream << ");\n";
    }
    node = node->Next();
  }
  return TRUE;
}

char *ClipsDatabase::GenerateType(ClipsExpr *expr, ClipsType type)
{
  static char buf[100];
  buf[0] = 0;
  switch (type)
  {
    case ClipsReal:
    {
      strcpy(buf, "float");
      break;
    }
    case ClipsInteger:
    {
      strcpy(buf, "long");
      break;
    }
    case ClipsString:
    case ClipsStringClass:
    case ClipsWord:
    {
      strcpy(buf, "wxString");
      break;
    }
    case ClipsList:
    {
      strcpy(buf, "ClipsExpr");
      break;
    }
/*
    case ClipsUser:
    {
      break;
    }
*/
    case ClipsNull:
    {
      strcpy(buf, "void");
      break;
    }
    default:
    {
      strcpy(buf, "void");
      break;
    }
  }
  return buf;
}


Bool ClipsDatabase::GenerateStrcat(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent, ostream& stream)
{
  ClipsExpr *operand = expr->Nth(1);
  if (expr->GetReturnType() == ClipsNull)
    Indent(stream, indent);
    
  stream << "(";
  while (operand)
  {
    GenerateExpression(func, operand, returnType, indent, stream);
    operand = operand->GetNext();
    if (operand) stream << " + ";
  }
  stream << ")";
  return TRUE;
}

Bool ClipsDatabase::GenerateSubstring(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent, ostream& stream)
{
  ClipsExpr *operand1 = expr->Nth(1);
  ClipsExpr *operand2 = expr->Nth(2);
  ClipsExpr *operand3 = expr->Nth(3);
  ClipsType typ;
  if (expr->GetReturnType() == ClipsNull)
    Indent(stream, indent);
    
  GenerateExpression(func, operand3, &typ, indent, stream);
  stream << ".Substring(";
  GenerateExpression(func, operand1, &typ, indent, stream);
  GenerateExpression(func, operand2, &typ, indent, stream);
  stream << ")";
  return TRUE;
}

Bool ClipsDatabase::GenerateUpcase(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent, ostream& stream)
{
  ClipsExpr *operand = expr->Nth(1);
  ClipsType typ;
  if (expr->GetReturnType() == ClipsNull)
    Indent(stream, indent);
    
  GenerateExpression(func, operand, &typ, indent, stream);
  stream << ".UpperCase()";
  return TRUE;
}

Bool ClipsDatabase::GenerateLowcase(ClipsConstruct *func, ClipsExpr *expr,
                               ClipsType *returnType, int indent, ostream& stream)
{
  ClipsExpr *operand = expr->Nth(1);
  ClipsType typ;
  if (expr->GetReturnType() == ClipsNull)
    Indent(stream, indent);
    
  GenerateExpression(func, operand, &typ, indent, stream);
  stream << ".LowerCase()";
  return TRUE;
}

Bool ClipsDatabase::GenerateMvAppend(ClipsConstruct *func, ClipsExpr *expr,
                             ClipsType *returnType, int indent, ostream& stream)
{
  ClipsExpr *operand = expr->Nth(1);
  if (expr->GetReturnType() == ClipsNull)
    Indent(stream, indent);
    
  stream << "MakeList(";
  while (operand)
  {
    stream << "new ClipsExpr(";
    GenerateExpression(func, operand, returnType, indent, stream);
    stream << "), ";
    operand = operand->GetNext();
  }
  stream << "NULL)";
  return TRUE;
}

Bool ClipsDatabase::GeneratePredicate(ClipsConstruct *func, ClipsExpr *expr,
                             ClipsType *returnType, int indent, ostream& stream)
{
  ClipsExpr *pred = expr->Nth(0);
  char *predName = pred->WordValue();
  ClipsExpr *operand = expr->Nth(1);

  expr->SetReturnType(ClipsInteger);
  *returnType = ClipsInteger;

  if (strcmp(predName, "numberp") == 0)
  {
    if (operand->GetReturnType() == ClipsInteger || operand->GetReturnType() == ClipsReal)
      stream << "TRUE";
    else
      stream << "FALSE";
  }
  else if (strcmp(predName, "floatp") == 0)
  {
    if (operand->GetReturnType() == ClipsReal)
      stream << "TRUE";
    else
      stream << "FALSE";
  }
  else if (strcmp(predName, "integerp") == 0)
  {
    if (operand->GetReturnType() == ClipsInteger)
      stream << "TRUE";
    else
      stream << "FALSE";
  }
  else if (strcmp(predName, "lexemep") == 0)
  {
    if (operand->GetReturnType() == ClipsString || operand->GetReturnType() == ClipsWord ||
        operand->GetReturnType() == ClipsStringClass)
      stream << "TRUE";
    else
      stream << "FALSE";
  }
  else if (strcmp(predName, "stringp") == 0)
  {
    if (operand->GetReturnType() == ClipsString || operand->GetReturnType() == ClipsWord ||
        operand->GetReturnType() == ClipsStringClass)
      stream << "TRUE";
    else
      stream << "FALSE";
  }
  else if (strcmp(predName, "floatp") == 0)
  {
    if (operand->GetReturnType() == ClipsReal)
      stream << "TRUE";
    else
      stream << "FALSE";
  }
  else if (strcmp(predName, "evenp") == 0)
  {
    if (operand->GetReturnType() == ClipsReal)
      stream << "TRUE";
    else
      stream << "FALSE";
  }
  else if (strcmp(predName, "multifieldp") == 0)
  {
    if (operand->GetReturnType() == ClipsList)
      stream << "TRUE";
    else
      stream << "FALSE";
  }
  
  return TRUE;
}

Bool ClipsDatabase::GenerateNth(ClipsConstruct *func, ClipsExpr *expr,
                             ClipsType *returnType, int indent, ostream& stream)
{
  char buf[100];
  ClipsExpr *listExpr = expr->Nth(2);
  ClipsExpr *operand = expr->Nth(1);

  ClipsType lType;
  if (expr->GetReturnType() == ClipsNull)
    Indent(stream, indent);
  GenerateExpression(func, listExpr, &lType, indent, stream);
  stream << ".";

  switch (expr->GetReturnType())
  {
    case ClipsReal:
    {
      stream << "NthReal(";
      break;
    }
    case ClipsInteger:
    {
      stream << "NthInteger(";
      break;
    }
    case ClipsString:
    case ClipsWord:
    case ClipsStringClass:
    {
      stream << "NthString(";
      break;
    }
    case ClipsList:
    {
      ClipsError("Cannot access a list member using nth.");
      break;
    }
    default:
    {
      sprintf(buf, "In function %s, could not resolve return type for nth.", func->GetName());
      ClipsError(buf);
      break;
    }
  }
  
  ClipsType nType;
  stream << "(";
  GenerateExpression(func, operand, &nType, indent, stream);
  stream << ")";
  stream << ")";

  return TRUE;
}


Bool ClipsDatabase::GenerateMember(ClipsConstruct *func, ClipsExpr *expr,
                             ClipsType *returnType, int indent, ostream& stream)
{
  ClipsExpr *operand1 = expr->Nth(1);
  ClipsExpr *operand2 = expr->Nth(2);
  ClipsType typ;
  if (expr->GetReturnType() == ClipsNull)
    Indent(stream, indent);
  GenerateExpression(func, operand2, &typ, indent, stream);
  stream << ".Member(";
  GenerateExpression(func, operand1, &typ, indent, stream);
  stream << ")";
  return TRUE;
}

Bool ClipsDatabase::GenerateLength(ClipsConstruct *func, ClipsExpr *expr,
                             ClipsType *returnType, int indent, ostream& stream)
{
  ClipsExpr *operand = expr->Nth(1);
  ClipsType typ;
  if (expr->GetReturnType() == ClipsNull)
    Indent(stream, indent);
  GenerateExpression(func, operand, &typ, indent, stream);
  if (operand->Type() == ClipsList)
  {
    stream << ".Number()";
  }
  else
  {
    stream << ".Length()";
  }
  *returnType = ClipsInteger;
  return TRUE;
}

Bool ClipsDatabase::GenerateFormat(ClipsConstruct *func, ClipsExpr *expr,
                             ClipsType *returnType, int indent, ostream& stream)
{
  ClipsExpr *operand1 = expr->Nth(1);
  ClipsExpr *operand2 = expr->Nth(2);
  ClipsExpr *operand3 = expr->Nth(3);
  ClipsType typ;

  if (expr->GetReturnType() == ClipsNull)
    Indent(stream, indent);
    
  stream << "ClipsFormat(";
  GenerateExpression(func, operand1, &typ, indent, stream);
  stream << ", TranslateFormat(";
  GenerateExpression(func, operand2, &typ, indent, stream);
  stream << ")";
  while (operand3)
  {
    stream << ",";
    GenerateExpression(func, operand3, &typ, indent, stream);
    if (operand3->GetReturnType() == ClipsWord || operand3->GetReturnType() == ClipsString ||
        operand3->GetReturnType() == ClipsStringClass)
      stream << ".GetData()";
    operand3 = operand3->GetNext();
  }
  stream << ")";
  *returnType = ClipsString;
  return TRUE;
}

Bool ClipsDatabase::GeneratePrintout(ClipsConstruct *func, ClipsExpr *expr,
                             ClipsType *returnType, int indent, ostream& stream)
{
  ClipsExpr *operand1 = expr->Nth(1);
  ClipsExpr *operand2 = expr->Nth(2);
  ClipsType typ;

  if (expr->GetReturnType() == ClipsNull)
    Indent(stream, indent);
    
  stream << "ClipsPrintout(";
  GenerateExpression(func, operand1, &typ, indent, stream);
  stream << ", MakeList(";
  while (operand2)
  {
    if (operand2->Type() == ClipsWord && (strcmp(operand2->WordValue(), "crlf") == 0))
      stream << "wxString(\"\\n\")";
    else
      GenerateExpression(func, operand2, &typ, indent, stream);
    operand2 = operand2->GetNext();
    if (operand2)
      stream << ",";
  }
  stream << ", NULL))";
  *returnType = ClipsNull;
  return TRUE;
}

Bool ClipsDatabase::GenerateReadline(ClipsConstruct *func, ClipsExpr *expr,
                             ClipsType *returnType, int indent, ostream& stream)
{
  ClipsExpr *operand = expr->Nth(1);
  ClipsType typ;

  if (expr->GetReturnType() == ClipsNull)
    Indent(stream, indent);

  if (!operand)
  {
    char buf[250];
    sprintf(buf, "In function %s, readline uses standard input.\nUsing wxcGetTextFromUser instead.", func->GetName());
    ClipsWarning(buf);
    stream << "wxString(wxcGetTextFromUser(\"Enter text\"))";
  }
  else
  {
    stream << "ClipsReadLine(";
    GenerateExpression(func, operand, &typ, indent, stream);
    stream << ")";
  }
  *returnType = ClipsString;
  return TRUE;
}

Bool ClipsDatabase::GenerateRead(ClipsConstruct *func, ClipsExpr *expr,
                             ClipsType *returnType, int indent, ostream& stream)
{
  char buf[300];
  ClipsExpr *operand = expr->Nth(1);
  ClipsType typ;

  if (expr->GetReturnType() == ClipsNull)
    Indent(stream, indent);

  if (!operand)
  {
    sprintf(buf, "In function %s, read uses standard input.\nUsing wxcGetTextFromUser instead.", func->GetName());
    ClipsWarning(buf);
    switch (expr->GetReturnType())
    {
      case ClipsReal:
      {
        stream << "atof(";
        break;
      }
      case ClipsInteger:
      {
        stream << "atod(";
        break;
      }
      case ClipsString:
      case ClipsWord:
      case ClipsUser:
      case ClipsStringClass:
      {
        stream << "wxString(";
        break;
      }
      default:
      {
        sprintf(buf, "In function %s, unresolved type for read. Use declare to resolve type.", func->GetName());
        ClipsError(buf);
        stream << "(";
        break;
      }
    }
    stream << "wxcGetTextFromUser(\"Enter text\"))";
  }
  else
  {
    switch (expr->GetReturnType())
    {
      case ClipsReal:
      {
        stream << "ClipsReadReal(";
        break;
      }
      case ClipsInteger:
      {
        stream << "ClipsReadInteger(";
        break;
      }
      case ClipsString:
      case ClipsWord:
      case ClipsUser:
      case ClipsStringClass:
      {
        stream << "ClipsReadString(";
        break;
      }
      default:
      {
        sprintf(buf, "In function %s, unresolved type for read. Use declare to resolve type.", func->GetName());
        ClipsError(buf);
        stream << "ClipsReadString(";
        break;
      }
    }

    GenerateExpression(func, operand, &typ, indent, stream);
    stream << ")";
  }
  return TRUE;
}

Bool ClipsDatabase::GenerateOpen(ClipsConstruct *func, ClipsExpr *expr,
                             ClipsType *returnType, int indent, ostream& stream)
{
  ClipsExpr *operand1 = expr->Nth(1);
  ClipsExpr *operand2 = expr->Nth(2);
  ClipsExpr *operand3 = expr->Nth(3);
  ClipsType typ;
  if (expr->GetReturnType() == ClipsNull)
    Indent(stream, indent);
    
  stream << "ClipsOpen(";
  GenerateExpression(func, operand1, &typ, indent, stream);
  stream << ", ";
  GenerateExpression(func, operand2, &typ, indent, stream);
  stream << ", ";
  if (operand3)
  {
    GenerateExpression(func, operand3, &typ, indent, stream);
  }
  else
    stream << "wxString(\"r\")";
    
  stream << ")";
  *returnType = ClipsInteger;
  return TRUE;
}

Bool ClipsDatabase::GenerateClose(ClipsConstruct *func, ClipsExpr *expr,
                             ClipsType *returnType, int indent, ostream& stream)
{
  ClipsExpr *operand1 = expr->Nth(1);
  ClipsType typ;
  if (expr->GetReturnType() == ClipsNull)
    Indent(stream, indent);
    
  stream << "ClipsClose(";
  if (operand1)
    GenerateExpression(func, operand1, &typ, indent, stream);
  else
    stream << "wxString(\"\")";
    
  stream << ")";
  *returnType = ClipsInteger;
  return TRUE;
}

Bool ClipsDatabase::GenerateDeleteSubseq(ClipsConstruct *func, ClipsExpr *expr,
                             ClipsType *returnType, int indent, ostream& stream)
{
  ClipsExpr *op = expr->Nth(0);
  char *opName = op->WordValue();
  ClipsExpr *mvExpr = expr->Nth(1);
  ClipsExpr *operand1 = expr->Nth(2);
  ClipsExpr *operand2 = expr->Nth(3);
  ClipsType typ;
  if (expr->GetReturnType() == ClipsNull)
    Indent(stream, indent);
    
  GenerateExpression(func, mvExpr, &typ, indent, stream);
  if (strcmp(opName, "delete") == 0 || strcmp(opName, "delete$") == 0)
    stream << ".DeleteRange(";
  else
    stream << ".Subseq(";
  GenerateExpression(func, operand1, &typ, indent, stream);
  stream << ",";
  GenerateExpression(func, operand2, &typ, indent, stream);
  stream << ")";
  *returnType = ClipsList;
  return TRUE;
}

Bool ClipsDatabase::GenerateExplode(ClipsConstruct *func, ClipsExpr *expr,
                             ClipsType *returnType, int indent, ostream& stream)
{
  ClipsExpr *operand = expr->Nth(1);
  ClipsType typ;
  if (expr->GetReturnType() == ClipsNull)
    Indent(stream, indent);
    
  stream << "StringExplode(";
  GenerateExpression(func, operand, &typ, indent, stream);
  stream << ")";
  *returnType = ClipsList;
  return TRUE;
}

Bool ClipsDatabase::GenerateImplode(ClipsConstruct *func, ClipsExpr *expr,
                             ClipsType *returnType, int indent, ostream& stream)
{
  ClipsExpr *operand = expr->Nth(1);
  ClipsType typ;
  if (expr->GetReturnType() == ClipsNull)
    Indent(stream, indent);
    
  GenerateExpression(func, operand, &typ, indent, stream);
  stream << ".Implode()";
  *returnType = ClipsStringClass;
  return TRUE;
}

Bool ClipsDatabase::GenerateInsert(ClipsConstruct *func, ClipsExpr *expr,
                             ClipsType *returnType, int indent, ostream& stream)
{
  ClipsExpr *mvExpr = expr->Nth(1);
  ClipsExpr *operand1 = expr->Nth(2);
  ClipsExpr *operand2 = expr->Nth(3);
  ClipsType typ;
  if (expr->GetReturnType() == ClipsNull)
    Indent(stream, indent);
    
  GenerateExpression(func, mvExpr, &typ, indent, stream);
  stream << ".Insert(";
  GenerateExpression(func, operand1, &typ, indent, stream);
  stream << ",";
  GenerateExpression(func, operand2, &typ, indent, stream);
  stream << ")";
  *returnType = ClipsList;
  return TRUE;
}

Bool ClipsDatabase::GenerateReplace(ClipsConstruct *func, ClipsExpr *expr,
                             ClipsType *returnType, int indent, ostream& stream)
{
  ClipsExpr *mvExpr = expr->Nth(1);
  ClipsExpr *operand1 = expr->Nth(2);
  ClipsExpr *operand2 = expr->Nth(3);
  ClipsExpr *operand3 = expr->Nth(4);
  ClipsType typ;
  if (expr->GetReturnType() == ClipsNull)
    Indent(stream, indent);
    
  GenerateExpression(func, mvExpr, &typ, indent, stream);
  stream << ".Replace(";
  GenerateExpression(func, operand1, &typ, indent, stream);
  stream << ",";
  GenerateExpression(func, operand2, &typ, indent, stream);
  stream << ",";
  GenerateExpression(func, operand3, &typ, indent, stream);
  stream << ")";
  *returnType = ClipsList;
  return TRUE;
}

Bool ClipsDatabase::GenerateFirstRest(ClipsConstruct *func, ClipsExpr *expr,
                             ClipsType *returnType, int indent, ostream& stream)
{
  ClipsExpr *opExpr = expr->Nth(0);
  char *opName = opExpr->WordValue();
  ClipsExpr *mvExpr = expr->Nth(1);
  ClipsType typ;
  if (expr->GetReturnType() == ClipsNull)
    Indent(stream, indent);
    
  GenerateExpression(func, mvExpr, &typ, indent, stream);
  if (strncmp(opName, "first", 5) == 0)
    stream << ".First()";
  else
    stream << ".Rest()";
  *returnType = ClipsList;
  return TRUE;
}

void Indent(ostream& stream, int indent)
{
  for (int i = 0; i < indent; i ++)
    stream << ' ';
}
