/*
 * File:	main.cc
 * Purpose:	Main driver for clips2c
 * Author:	Julian Smart
 * Created:	1993
 * Updated:	
 * Copyright:	(c) 1993, AIAI, University of Edinburgh
 */

static const char sccsid[] = "%W% %G%";

#include <wx.h>

#ifndef NO_GUI
#include <wx_help.h>
#include <wx_timer.h>
#endif
#ifdef NO_GUI
#include <iostream.h>
#endif

#include "clips2c.h"
#include "wxstring.h"
#include "main.h"

Bool errorHandler(int errorType, char *msg);

ClipsDatabase database(errorHandler);

#if (defined(wx_x) && !defined(NO_GUI))
#include "aiai.xbm"
#endif

const float versionNo = 1.1;
Bool isInteractive = FALSE;
int passNumber = 1;
Bool isSync = FALSE;
Bool runTwice = FALSE;
int generateMode = MODE_PLAIN_CLIPS;

#ifndef NO_GUI
wxHelpInstance *HelpInstance = NULL;
#endif

char *InputFile = NULL;
char *OutputFile = NULL;
char *FileRoot = NULL;

// int BufSize = 40;             // Size of buffer in K

Bool Go(void);
void ShowOptions(void);

#ifdef NO_GUI
int main(int argc, char **argv)
#else
wxMenuBar *menuBar = NULL;
MyFrame *frame = NULL;

// This statement initializes the whole application and calls OnInit
MyApp myApp;

// `Main program' equivalent, creating windows and returning main app frame
wxFrame *MyApp::OnInit(void)
#endif
{
  InitializeClips2C(generateMode);
  
  FileRoot = new char[300];

  int n = 1;
  
  // Read input/output files
  if (argc > 1)
  {
    if (argv[1][0] != '-')
    {
      InputFile = argv[1];
      n ++;

      if (argc > 2)
      {
        if (argv[2][0] != '-')
        {
          OutputFile = argv[2];
          n ++;
        }
      }
    }
  }

#ifdef NO_GUI
  if (!InputFile || !OutputFile)
  {
    cout << "Clips2C: input or output file is missing.\n";
    ShowOptions();
    exit(1);
  }
#endif

  for (int i = n; i < argc;)
  {
    if (strcmp(argv[i], "-sync") == 0)  // Don't yield
    {
      i ++;
      isSync = TRUE;
    }
#ifndef NO_GUI
    else if (strcmp(argv[i], "-interactive") == 0)
    {
      i ++;
      isInteractive = TRUE;
    }
#endif
    else if (strcmp(argv[i], "-clips") == 0) // Plain CLIPS mode
    {
      i ++;
      generateMode = MODE_PLAIN_CLIPS;
    }
    else if (strcmp(argv[i], "-wxclips") == 0) // wxCLIPS mode
    {
      i ++;
      generateMode = MODE_WXCLIPS;
    }
    else if (strcmp(argv[i], "-hardy") == 0) // HARDY mode
    {
      i ++;
      generateMode = MODE_HARDY;
    }
/*
    else if (strcmp(argv[i], "-twice") == 0)
    {
      i ++;
      runTwice = TRUE;
    }
*/
    else
    {
      char buf[100];
      sprintf(buf, "Invalid switch %s.\n", argv[i]);
      OnClipsError(buf);
      i++;
#ifdef NO_GUI
      ShowOptions();
      exit(1);
#endif
    }
  }

#ifndef NO_GUI
  if (!InputFile && !OutputFile)
    isInteractive = TRUE;

  char buf[100];
#ifdef NO_GUI
  switch (generateMode)
  {
    case MODE_PLAIN_CLIPS:
    {
      OnClipsInform("Plain CLIPS mode.");
      break;
    }
    case MODE_WXCLIPS:
    {
      OnClipsInform("wxCLIPS mode.");
      break;
    }
    case MODE_HARDY:
    {
      OnClipsInform("HARDY mode.");
      break;
    }
    default:
      break;
  }
#endif
  if (isInteractive)
  {
    // Create the main frame window
    frame = new MyFrame(NULL, "CLIPS2C", 0, 0, 400, 300);
    frame->CreateStatusLine(2);
    // Give it an icon
#ifdef wx_msw
    wxIcon *icon = new wxIcon("clips2c_icn");
#endif
#ifdef wx_x
    wxIcon *icon = new wxIcon(aiai_bits, aiai_width, aiai_height);
#endif

    frame->SetIcon(icon);

//    frame->SetStatusText(buf, 1);

    if (InputFile)
    {
      sprintf(buf, "CLIPS2C [%s]", FileNameFromPath(InputFile));
      frame->SetTitle(buf);
    }

    // Make a menubar
    wxMenu *file_menu = new wxMenu;
    file_menu->Append(CLIPS_GO, "&Go",                        "Run converter");
    file_menu->Append(CLIPS_SET_INPUT, "Set &Input File",     "Set the CLIPS input file");
    file_menu->Append(CLIPS_SET_OUTPUT, "Set &Output File",   "Set the output file");
    file_menu->AppendSeparator();
    file_menu->Append(CLIPS_VIEW_LATEX, "View &CLIPS File",   "View the CLIPS input file");
    file_menu->Append(CLIPS_VIEW_OUTPUT, "View Output &File", "View the C++ output file");
    file_menu->Append(CLIPS_SAVE_FILE, "&Save file", "Save current text into file");
    file_menu->AppendSeparator();
    file_menu->Append(CLIPS_QUIT, "E&xit",                    "Exit Clips2C");

    wxMenu *mode_menu = new wxMenu;

    mode_menu->Append(CLIPS_MODE_PLAIN, "Interpret plain &CLIPS",   "Interpret plain CLIPS");
    mode_menu->Append(CLIPS_MODE_WXCLIPS, "Interpret &wxCLIPS", "Interpret wxCLIPS");
    mode_menu->Append(CLIPS_MODE_HARDY, "Interpret &HARDY",        "Interpret wxCLIPS/HARDY");

    wxMenu *help_menu = new wxMenu;

    help_menu->Append(CLIPS_HELP, "&Help", "CLIPS2C Contents Page");
    help_menu->Append(CLIPS_ABOUT, "&About CLIPS2C", "About CLIPS2C");

    menuBar = new wxMenuBar;
    menuBar->Append(file_menu, "&File");
    menuBar->Append(mode_menu, "&Conversion Mode");
    menuBar->Append(help_menu, "&Help");

    frame->SetMenuBar(menuBar);
    frame->textWindow = new wxTextWindow(frame);

    (*frame->textWindow) << "Welcome to Julian Smart's CLIPS to C++ converter.\n\n";
    ShowOptions();

    switch (generateMode)
    {
      case MODE_PLAIN_CLIPS:
      {
        frame->SetStatusText("CLIPS mode", 1);
        break;
      }
      case MODE_WXCLIPS:
      {
        frame->SetStatusText("wxCLIPS mode", 1);
        break;
      }
      case MODE_HARDY:
      {
        frame->SetStatusText("HARDY mode", 1);
        break;
      }
      default:
        break;
    }

    HelpInstance = new wxHelpInstance(TRUE);
    HelpInstance->Initialize("clips2c");

    frame->Show(TRUE);
    return frame;
  }
  else
#endif // NO_GUI
  {
    Go();
    if (runTwice) Go();
#ifdef NO_GUI
    return 0;
#else
    return NULL;
#endif
  }

#ifndef NO_GUI
  // Return the main frame window
  return frame;
#else
  return 0;
#endif
}

void ShowOptions(void)
{
    OnClipsInform("Usage: clips2c [input] [output] [switches]\n");
    OnClipsInform("where valid switches are");
    OnClipsInform("    -interactive");
    OnClipsInform("    -clips");
    OnClipsInform("    -wxclips");
    OnClipsInform("    -hardy\n");
}

#ifndef NO_GUI
// My frame constructor
MyFrame::MyFrame(wxFrame *frame, char *title, int x, int y, int w, int h):
  wxFrame(frame, title, x, y, w, h)
{}

// Intercept menu commands
void MyFrame::OnMenuCommand(int id)
{
  switch (id) {
    case CLIPS_QUIT:
      delete this;
      break;
    case CLIPS_GO:
    {
      menuBar->EnableTop(0, FALSE);
      menuBar->EnableTop(1, FALSE);
      menuBar->EnableTop(2, FALSE);
      textWindow->Clear();
      wxYield();
      Go();

      if (runTwice)
      {
        wxYield();
        Go();
      }
      menuBar->EnableTop(0, TRUE);
      menuBar->EnableTop(1, TRUE);
      menuBar->EnableTop(2, TRUE);
      break;
    }
    case CLIPS_SET_INPUT:
    {
      ChooseInputFile(TRUE);
      break;
    }
    case CLIPS_SET_OUTPUT:
    {
      ChooseOutputFile(TRUE);
      break;
    }
    case CLIPS_SAVE_FILE:
    {
      char *s = wxFileSelector("Save text to file", NULL, NULL, "txt", "*.txt");
      if (s)
      {
        textWindow->SaveFile(s);
        char buf[350];
        sprintf(buf, "Saved text to %s", s);
        frame->SetStatusText(buf, 0);
      }
      break;
    }
    case CLIPS_VIEW_OUTPUT:
    {
      ChooseOutputFile();
      if (OutputFile && FileExists(OutputFile))
      {
        textWindow->LoadFile(OutputFile);
        char buf[300];
        sprintf(buf, "CLIPS2C [%s]", FileNameFromPath(OutputFile));
        frame->SetTitle(buf);
      }
      break;
    }
    case CLIPS_VIEW_LATEX:
    {
      ChooseInputFile();
      if (InputFile && FileExists(InputFile))
      {
        textWindow->LoadFile(InputFile);
        char buf[300];
        sprintf(buf, "CLIPS2C [%s]", FileNameFromPath(InputFile));
        frame->SetTitle(buf);
      }
      break;
    }
    case CLIPS_MODE_PLAIN:
    {
      generateMode = MODE_PLAIN_CLIPS;
      InitializeClips2C(generateMode);
      SetStatusText("CLIPS mode", 1);
      break;
    }
    case CLIPS_MODE_WXCLIPS:
    {
      generateMode = MODE_WXCLIPS;
      InitializeClips2C(generateMode);
      SetStatusText("wxCLIPS mode", 1);
      break;
    }
    case CLIPS_MODE_HARDY:
    {
      generateMode = MODE_HARDY;
      InitializeClips2C(generateMode);
      SetStatusText("HARDY mode", 1);
      break;
    }
    case CLIPS_HELP:
    {
      HelpInstance->LoadFile();
      HelpInstance->DisplayContents();
      break;
    }
    case CLIPS_ABOUT:
    {
      char buf[300];
      sprintf(buf, "CLIPS2C Version %.2f\nCLIPS function to C++ conversion\n\n(c) Julian Smart 1994", versionNo);
      wxMessageBox(buf, "About CLIPS2C");
      break;
    }
  }
}

void ChooseInputFile(Bool force)
{
  if (force || !InputFile)
  {
    char *s = wxFileSelector("Choose CLIPS input file", PathOnly(InputFile), FileNameFromPath(InputFile), "clp", "*.clp");
    if (s)
    {
      passNumber = 1;
      char buf[300];
      InputFile = copystring(s);
      sprintf(buf, "CLIPS2C [%s]", FileNameFromPath(InputFile));
      frame->SetTitle(buf);
      OutputFile = NULL;
    }
  }
}

void ChooseOutputFile(Bool force)
{
  char extensionBuf[10];
  char wildBuf[10];
  strcpy(wildBuf, "*.");
  char *path = NULL;
  if (OutputFile)
    path = PathOnly(OutputFile);
  else if (InputFile)
    path = PathOnly(InputFile);
    
  strcpy(extensionBuf, "cc");
  strcat(wildBuf, "cc");

  if (force || !OutputFile)
  {
    char *s = wxFileSelector("Choose output file", path, FileNameFromPath(OutputFile),
                   extensionBuf, wildBuf);
    if (s)
      OutputFile = copystring(s);
  }
}
#endif

Bool Go(void)
{
  char buf[300];
#ifndef NO_GUI
  ChooseInputFile();
  ChooseOutputFile();
#endif

  if (!InputFile || !OutputFile)
    return FALSE;

#ifndef NO_GUI
  if (isInteractive)
  {
    sprintf(buf, "CLIPS2C [%s]", FileNameFromPath(InputFile));
    frame->SetTitle(buf);
  }

  wxStartTimer();
#endif

  if (InputFile && OutputFile)
  {
    if (!FileExists(InputFile))
    {
      OnClipsError("Cannot open input file!");
      return FALSE;
    }
#ifndef NO_GUI
    if (isInteractive)
    {
      char buf[50];
      sprintf(buf, "Working, pass %d...", passNumber);
      frame->SetStatusText(buf);
    }
#endif
    OnClipsInform("Reading CLIPS file...");
    database.Clear();
    database.ReadClips(InputFile);
    sprintf(buf, "%d objects read in, generating...", database.Number());
    OnClipsInform(buf);
    database.GenerateC(OutputFile);
  }

#ifndef NO_GUI
  long tim = wxGetElapsedTime();
  sprintf(buf, "Finished in %ld seconds.", (long)(tim/1000.0));
  OnClipsInform(buf);
  if (isInteractive)
  {
    sprintf(buf, "Done, %d %s.", passNumber, (passNumber > 1) ? "passes" : "pass");
    frame->SetStatusText(buf);
  }
#else
  sprintf(buf, "Done, %d %s.", passNumber, (passNumber > 1) ? "passes" : "pass");
  OnClipsInform(buf);
#endif
  passNumber ++;
  return TRUE;
}

void OnClipsError(char *msg)
{
#ifdef NO_GUI
  cerr << "Error: " << msg << "\n";
#else
  if (isInteractive)
    (*frame->textWindow) << "Error: " << msg << "\n";
  else
#ifdef wx_x
    cerr << "Error: " << msg << "\n";
#endif
#ifdef wx_msw
    wxError(msg);
#endif
  wxYield();
#endif // NO_GUI
}

void OnClipsWarning(char *msg)
{
#ifdef NO_GUI
  cerr << "Warning: " << msg << "\n";
#else
  if (isInteractive)
    (*frame->textWindow) << "Warning: " << msg << "\n";
  else
#ifdef wx_x
    cerr << "Error: " << msg << "\n";
#endif
#ifdef wx_msw
    wxError(msg);
#endif
  wxYield();
#endif // NO_GUI
}

void OnClipsInform(char *msg)
{
#ifdef NO_GUI
  cout << msg << "\n";
#else
  if (isInteractive)
    (*frame->textWindow) << msg << "\n";
  else
#ifdef wx_x
    cerr << msg << "\n";
#endif
#ifdef wx_msw
    {}
#endif
  if (isInteractive)
    wxYield();
#endif // NO_GUI
}


Bool errorHandler(int errorType, char *msg)
{
#ifdef NO_GUI
  cerr << "Error during loading: " << msg << "\n";
#else
  char buf[500];
  sprintf(buf, "Error during loading: %s", msg);
  wxMessageBox(buf, "Error");
#endif
  return FALSE;
}

void ClipsErrorFunction(char *buffer)
{
//  if (frame)
//    (*frame->textWindow) << buffer;
}
